@@ -1,4 +1,4 @@
-# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2014 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 from __future__ import absolute_import
@@ -11,34 +11,40 @@ from collections import OrderedDict, defaultdict
 
 from pex import dist_metadata, pex_warnings, targets
 from pex.common import pluralize
+from pex.dependency_configuration import DependencyConfiguration
+from pex.dist_metadata import Distribution, Requirement, is_wheel
 from pex.fingerprinted_distribution import FingerprintedDistribution
 from pex.inherit_path import InheritPath
-from pex.layout import maybe_install
+from pex.interpreter import PythonInterpreter
+from pex.layout import ensure_installed, identify_layout
 from pex.orderedset import OrderedSet
 from pex.pep_425 import TagRank
-from pex.pep_503 import ProjectName, distribution_satisfies_requirement
+from pex.pep_503 import ProjectName
 from pex.pex_info import PexInfo
 from pex.targets import Target
-from pex.third_party.packaging import specifiers, tags
-from pex.third_party.pkg_resources import Distribution, Requirement
+from pex.third_party.packaging import specifiers
+from pex.third_party.packaging.tags import Tag
 from pex.tracer import TRACER
 from pex.typing import TYPE_CHECKING
-from pex.util import DistributionHelper
 
 if TYPE_CHECKING:
     from typing import (
         DefaultDict,
+        Dict,
         FrozenSet,
         Iterable,
         Iterator,
         List,
         MutableMapping,
         Optional,
+        Text,
         Tuple,
         Union,
     )
 
     import attr  # vendor:skip
+
+    from pex.pep_427 import InstallableType  # noqa
 else:
     from pex.third_party import attr
 
@@ -52,9 +58,12 @@ def _import_pkg_resources():
         from pex import third_party
 
         third_party.install(expose=["setuptools"])
-        import pkg_resources  # vendor:skip
+        try:
+            import pkg_resources  # vendor:skip
 
-        return pkg_resources, True
+            return pkg_resources, True
+        except ImportError:
+            return None, False
 
 
 @attr.s(frozen=True)
@@ -98,7 +107,7 @@ class _RankedDistribution(object):
 
     def satisfies(self, requirement):
         # type: (Requirement) -> bool
-        return distribution_satisfies_requirement(self.distribution, requirement)
+        return self.distribution in requirement
 
 
 @attr.s(frozen=True)
@@ -131,17 +140,18 @@ class _InvalidWheelName(_UnrankedDistribution):
 
 @attr.s(frozen=True)
 class _TagMismatch(_UnrankedDistribution):
-    wheel_tags = attr.ib()  # type: FrozenSet[tags.Tag]
+    wheel_tags = attr.ib()  # type: Iterable[Tag]
 
     def render_message(self, target):
         # type: (Target) -> str
         return (
             "The wheel tags for {dist} are {wheel_tags} which do not match the supported tags of "
-            "{target}:\n{supported_tags}".format(
+            "{target}:\n{tag}\n... {count} more ...".format(
                 dist=self.dist,
                 wheel_tags=", ".join(map(str, self.wheel_tags)),
                 target=target,
-                supported_tags="\n".join(map(str, target.supported_tags)),
+                tag=target.supported_tags[0],
+                count=len(target.supported_tags) - 1,
             )
         )
 
@@ -167,6 +177,10 @@ class _QualifiedRequirement(object):
     requirement = attr.ib()  # type: Requirement
     required = attr.ib(default=True)  # type: bool
 
+    def with_extras(self, extras):
+        # type: (FrozenSet[str]) -> _QualifiedRequirement
+        return attr.evolve(self, requirement=attr.evolve(self.requirement, extras=extras))
+
 
 @attr.s(frozen=True)
 class _DistributionNotFound(object):
@@ -183,12 +197,13 @@ class ResolveError(Exception):
 
 
 @attr.s(frozen=True)
-class _RequirementKey(ProjectName):
+class _RequirementKey(object):
     @classmethod
     def create(cls, requirement):
         # type: (Requirement) -> _RequirementKey
-        return cls(requirement, frozenset(requirement.extras))
+        return cls(ProjectName(requirement.name), frozenset(requirement.extras))
 
+    project_name = attr.ib()  # type: ProjectName
     extras = attr.ib()  # type: FrozenSet[str]
 
     def satisfied_keys(self):
@@ -208,6 +223,8 @@ class _RequirementKey(ProjectName):
 
 
 class PEXEnvironment(object):
+    _CACHE = {}  # type: Dict[Tuple[str, str, Target], PEXEnvironment]
+
     @classmethod
     def mount(
         cls,
@@ -216,29 +233,37 @@ class PEXEnvironment(object):
         target=None,  # type: Optional[Target]
     ):
         # type: (...) -> PEXEnvironment
+        pex_file = os.path.realpath(pex)
         if not pex_info:
-            pex_info = PexInfo.from_pex(pex)
+            pex_info = PexInfo.from_pex(pex_file)
             pex_info.update(PexInfo.from_env())
         pex_hash = pex_info.pex_hash
         if pex_hash is None:
             raise AssertionError(
                 "There was no pex_hash stored in {} for {}.".format(PexInfo.PATH, pex)
             )
-        pex_root = pex_info.pex_root
-        pex = maybe_install(pex=pex, pex_root=pex_root, pex_hash=pex_hash) or pex
         target = target or targets.current()
-        return cls(pex=pex, pex_info=pex_info, target=target)
+        key = (pex_file, pex_hash, target)
+        mounted = cls._CACHE.get(key)
+        if mounted is None:
+            pex_root = pex_info.pex_root
+            installed_pex = ensure_installed(pex=pex, pex_root=pex_root, pex_hash=pex_hash)
+            mounted = cls(pex=installed_pex, pex_info=pex_info, target=target, source_pex=pex)
+            cls._CACHE[key] = mounted
+        return mounted
 
     def __init__(
         self,
         pex,  # type: str
         pex_info=None,  # type: Optional[PexInfo]
         target=None,  # type: Optional[Target]
+        source_pex=None,  # type: Optional[str]
     ):
         # type: (...) -> None
         self._pex = os.path.realpath(pex)
         self._pex_info = pex_info or PexInfo.from_pex(pex)
         self._target = target or targets.current()
+        self._source_pex = os.path.realpath(source_pex) if source_pex else None
 
         self._available_ranked_dists_by_project_name = defaultdict(
             list
@@ -254,16 +279,51 @@ class PEXEnvironment(object):
         # type: () -> str
         return self._pex
 
-    def _load_internal_cache(self):
-        # type: () -> Iterator[FingerprintedDistribution]
-        internal_cache = os.path.join(self._pex, self._pex_info.internal_cache)
-        with TRACER.timed("Searching dependency cache: %s" % internal_cache, V=2):
-            for distribution_name, fingerprint in self._pex_info.distributions.items():
-                dist_path = os.path.join(internal_cache, distribution_name)
-                yield FingerprintedDistribution(
-                    distribution=DistributionHelper.distribution_from_path(dist_path),
-                    fingerprint=fingerprint,
+    @property
+    def pex_info(self):
+        # type: () -> PexInfo
+        return self._pex_info
+
+    @property
+    def source_pex(self):
+        # type: () -> str
+        return self._source_pex or self._pex
+
+    def iter_distributions(self, result_type_wheel_file=False):
+        # type: (bool) -> Iterator[FingerprintedDistribution]
+        if result_type_wheel_file:
+            if not self._pex_info.deps_are_wheel_files:
+                raise ResolveError(
+                    "Cannot resolve .whl files from PEX at {pex}; its dependencies are in the "
+                    "form of pre-installed wheel chroots.".format(pex=self.source_pex)
                 )
+            with TRACER.timed(
+                "Searching dependency cache: {cache}".format(
+                    cache=os.path.join(self.source_pex, self._pex_info.internal_cache)
+                ),
+                V=2,
+            ):
+                with identify_layout(self.source_pex) as layout:
+                    for distribution_name, fingerprint in self._pex_info.distributions.items():
+                        yield FingerprintedDistribution(
+                            distribution=Distribution.load(
+                                layout.wheel_file_path(
+                                    (self._pex_info.internal_cache, distribution_name)
+                                )
+                            ),
+                            fingerprint=fingerprint,
+                        )
+        else:
+            internal_cache = os.path.join(self._pex, self._pex_info.internal_cache)
+            with TRACER.timed(
+                "Searching dependency cache: {cache}".format(cache=internal_cache), V=2
+            ):
+                for distribution_name, fingerprint in self._pex_info.distributions.items():
+                    dist_path = os.path.join(internal_cache, distribution_name)
+                    yield FingerprintedDistribution(
+                        distribution=Distribution.load(dist_path),
+                        fingerprint=fingerprint,
+                    )
 
     def _update_candidate_distributions(self, distribution_iter):
         # type: (Iterable[FingerprintedDistribution]) -> None
@@ -278,35 +338,26 @@ class PEXEnvironment(object):
 
     def _can_add(self, fingerprinted_dist):
         # type: (FingerprintedDistribution) -> Union[_RankedDistribution, _UnrankedDistribution]
-        filename, ext = os.path.splitext(os.path.basename(fingerprinted_dist.location))
-        if ext.lower() != ".whl":
+        filename = os.path.basename(fingerprinted_dist.location)
+        if not is_wheel(filename):
             # This supports resolving pex's own vendored distributions which are vendored in a
             # directory with the project name (`pip/` for pip) and not the corresponding wheel name
             # (`pip-19.3.1-py2.py3-none-any.whl/` for pip). Pex only vendors universal wheels for
             # all platforms it supports at buildtime and runtime so this is always safe.
             return _RankedDistribution.highest_rank(fingerprinted_dist)
 
-        # Wheel filename format: https://www.python.org/dev/peps/pep-0427/#file-name-convention
-        # `{distribution}-{version}(-{build tag})?-{python tag}-{abi tag}-{platform tag}.whl`
-        wheel_components = filename.split("-")
-        if len(wheel_components) < 3:
+        try:
+            wheel_eval = self._target.wheel_applies(fingerprinted_dist.distribution)
+        except ValueError:
             return _InvalidWheelName(fingerprinted_dist, filename)
 
-        # `{python tag}-{abi tag}-{platform tag}`
-        wheel_tags = tags.parse_tag("-".join(wheel_components[-3:]))
-        # There will be multiple parsed tags for compressed tag sets. Ensure we grab the parsed tag
-        # with highest rank from that expanded set.
-        best_match = self._target.supported_tags.best_match(wheel_tags)
-        if best_match is None:
-            return _TagMismatch(fingerprinted_dist, wheel_tags)
+        if not wheel_eval.best_match:
+            return _TagMismatch(fingerprinted_dist, wheel_eval.tags)
+        if not wheel_eval.applies:
+            assert wheel_eval.requires_python
+            return _PythonRequiresMismatch(fingerprinted_dist, wheel_eval.requires_python)
 
-        python_requires = dist_metadata.requires_python(fingerprinted_dist.distribution)
-        if python_requires and not self._target.requires_python_applies(
-            python_requires, source=fingerprinted_dist.distribution.as_requirement()
-        ):
-            return _PythonRequiresMismatch(fingerprinted_dist, python_requires)
-
-        return _RankedDistribution(best_match.rank, fingerprinted_dist)
+        return _RankedDistribution(wheel_eval.best_match.rank, fingerprinted_dist)
 
     def activate(self):
         # type: () -> Iterable[Distribution]
@@ -318,7 +369,7 @@ class PEXEnvironment(object):
     def _evaluate_marker(
         self,
         requirement,  # type: Requirement
-        extras=None,  # type: Optional[Tuple[str, ...]]
+        extras=(),  # type: Iterable[str]
     ):
         # type: (...) -> bool
         applies = self._target.requirement_applies(requirement, extras=extras)
@@ -334,11 +385,23 @@ class PEXEnvironment(object):
     def _resolve_requirement(
         self,
         requirement,  # type: Requirement
+        dependency_configuration,  # type: DependencyConfiguration
         resolved_dists_by_key,  # type: MutableMapping[_RequirementKey, FingerprintedDistribution]
         required,  # type: bool
         required_by=None,  # type: Optional[Distribution]
     ):
         # type: (...) -> Iterator[_DistributionNotFound]
+
+        excluded_by = dependency_configuration.excluded_by(requirement)
+        if excluded_by:
+            TRACER.log(
+                "Skipping resolving {requirement}: excluded by {excludes}".format(
+                    requirement=requirement,
+                    excludes=" and ".join(map(str, excluded_by)),
+                )
+            )
+            return
+
         requirement_key = _RequirementKey.create(requirement)
         if requirement_key in resolved_dists_by_key:
             return
@@ -346,7 +409,7 @@ class PEXEnvironment(object):
         available_distributions = [
             ranked_dist
             for ranked_dist in self._available_ranked_dists_by_project_name[
-                ProjectName(requirement)
+                requirement.project_name
             ]
             if ranked_dist.satisfies(requirement)
         ]
@@ -373,6 +436,17 @@ class PEXEnvironment(object):
         )
 
         for dep_requirement in dist_metadata.requires_dists(resolved_distribution.distribution):
+            override = dependency_configuration.overridden_by(dep_requirement, target=self._target)
+            if override:
+                TRACER.log(
+                    "Resolving {override}: overrides {requirement} from {dist}".format(
+                        override=override,
+                        requirement=dep_requirement,
+                        dist=os.path.basename(resolved_distribution.distribution.location),
+                    )
+                )
+                dep_requirement = override
+
             # A note regarding extras and why they're passed down one level (we don't pass / use
             # dep_requirement.extras for example):
             #
@@ -395,14 +469,19 @@ class PEXEnvironment(object):
 
             for not_found in self._resolve_requirement(
                 dep_requirement,
+                dependency_configuration,
                 resolved_dists_by_key,
                 required,
-                required_by=resolved_distribution,
+                required_by=resolved_distribution.distribution,
             ):
                 yield not_found
 
-    def _root_requirements_iter(self, reqs):
-        # type: (Iterable[Requirement]) -> Iterator[QualifiedRequirementOrNotFound]
+    def _root_requirements_iter(
+        self,
+        reqs,  # type: Iterable[Requirement]
+        dependency_configuration,  # type: DependencyConfiguration
+    ):
+        # type: (...) -> Iterator[QualifiedRequirementOrNotFound]
 
         # We want to pick one requirement for each key (required project) to then resolve
         # recursively.
@@ -419,10 +498,20 @@ class PEXEnvironment(object):
             OrderedDict()
         )  # type: OrderedDict[ProjectName, List[_QualifiedRequirement]]
         for req in reqs:
+            excluded_by = dependency_configuration.excluded_by(req)
+            if excluded_by:
+                TRACER.log(
+                    "Skipping resolving {requirement}: excluded by {excludes}".format(
+                        requirement=req,
+                        excludes=" and ".join(map(str, excluded_by)),
+                    )
+                )
+                continue
+
             required = self._evaluate_marker(req)
             if not required:
                 continue
-            project_name = ProjectName(req)
+            project_name = req.project_name
             requirements = qualified_reqs_by_project_name.get(project_name)
             if requirements is None:
                 qualified_reqs_by_project_name[project_name] = requirements = []
@@ -437,25 +526,24 @@ class PEXEnvironment(object):
                 # We've winnowed down reqs_by_key to just those requirements whose environment
                 # markers apply; so, we should always have an available distribution.
                 message = (
-                    "A distribution for {project_name} could not be resolved in this "
-                    "environment.".format(project_name=project_name)
+                    "A distribution for {project_name} could not be resolved for {target}.".format(
+                        project_name=project_name, target=self._target
+                    )
                 )
                 unavailable_dists = self._unavailable_dists_by_project_name.get(project_name)
                 if unavailable_dists:
-                    message += (
-                        "Found {count} {distributions} for {project_name} that do not apply:\n"
-                        "{unavailable_dists}".format(
-                            count=len(unavailable_dists),
-                            distributions=pluralize(unavailable_dists, "distribution"),
-                            project_name=project_name,
-                            unavailable_dists="\n".join(
-                                "{index}.) {message}".format(
-                                    index=index,
-                                    message=unavailable_dist.render_message(self._target),
-                                )
-                                for index, unavailable_dist in enumerate(unavailable_dists, start=1)
-                            ),
-                        )
+                    message += "\nFound {count} {distributions} for {project_name} that {does} not apply:\n" "{unavailable_dists}".format(
+                        count=len(unavailable_dists),
+                        distributions=pluralize(unavailable_dists, "distribution"),
+                        project_name=project_name,
+                        does="does" if len(unavailable_dists) == 1 else "do",
+                        unavailable_dists="\n".join(
+                            "{index}.) {message}".format(
+                                index=index,
+                                message=unavailable_dist.render_message(self._target),
+                            )
+                            for index, unavailable_dist in enumerate(unavailable_dists, start=1)
+                        ),
                     )
                 raise ResolveError(message)
             candidates = [
@@ -484,22 +572,51 @@ class PEXEnvironment(object):
                     ),
                     V=9,
                 )
-            yield qualified_requirement
+
+            # We may have had multiple requirements that select the winning candidate distribution.
+            # For example, say we're a Python 3.10 interpreter and the root requirements are
+            # `"foo[bar]; python_version < '3.11'" "foo[baz]==1.2.3"`. In that case, we want to
+            # ensure that for whichever ~random requirement we selected as a representative we
+            # gather all extras across the candidate requirements to make sure all requested extras
+            # are grafted in to the resolve.
+            yield qualified_requirement.with_extras(
+                frozenset(
+                    itertools.chain.from_iterable(
+                        candidate[1].requirement.extras for candidate in candidates
+                    )
+                )
+            )
 
     def resolve(self):
         # type: () -> Iterable[Distribution]
         if self._resolved_dists is None:
             all_reqs = [Requirement.parse(req) for req in self._pex_info.requirements]
+            dependency_configuration = DependencyConfiguration.from_pex_info(self._pex_info)
             self._resolved_dists = tuple(
                 fingerprinted_distribution.distribution
-                for fingerprinted_distribution in self.resolve_dists(all_reqs)
+                for fingerprinted_distribution in self.resolve_dists(
+                    all_reqs, dependency_configuration=dependency_configuration
+                )
             )
         return self._resolved_dists
 
-    def resolve_dists(self, reqs):
-        # type: (Iterable[Requirement]) -> Iterable[FingerprintedDistribution]
+    def resolve_dists(
+        self,
+        reqs,  # type: Iterable[Requirement]
+        dependency_configuration=DependencyConfiguration(),  # type: DependencyConfiguration
+        result_type=None,  # type: Optional[InstallableType.Value]
+    ):
+        # type: (...) -> Iterable[FingerprintedDistribution]
 
-        self._update_candidate_distributions(self._load_internal_cache())
+        result_type_wheel_file = False
+        if result_type is not None:
+            from pex.pep_427 import InstallableType
+
+            result_type_wheel_file = result_type is InstallableType.WHEEL_FILE
+
+        self._update_candidate_distributions(
+            self.iter_distributions(result_type_wheel_file=result_type_wheel_file)
+        )
 
         unresolved_reqs = OrderedDict()  # type: OrderedDict[Requirement, OrderedSet]
 
@@ -516,7 +633,9 @@ class PEXEnvironment(object):
         resolved_dists_by_key = (
             OrderedDict()
         )  # type: OrderedDict[_RequirementKey, FingerprintedDistribution]
-        for qualified_req_or_not_found in self._root_requirements_iter(reqs):
+        for qualified_req_or_not_found in self._root_requirements_iter(
+            reqs, dependency_configuration
+        ):
             if isinstance(qualified_req_or_not_found, _DistributionNotFound):
                 record_unresolved(qualified_req_or_not_found)
                 continue
@@ -524,6 +643,7 @@ class PEXEnvironment(object):
             with TRACER.timed("Resolving {}".format(qualified_req_or_not_found.requirement), V=2):
                 for not_found in self._resolve_requirement(
                     requirement=qualified_req_or_not_found.requirement,
+                    dependency_configuration=dependency_configuration,
                     required=qualified_req_or_not_found.required,
                     resolved_dists_by_key=resolved_dists_by_key,
                 ):
@@ -550,7 +670,7 @@ class PEXEnvironment(object):
                             requirers="\n      ".join(map(str, requirers))
                         )
                     contains = self._available_ranked_dists_by_project_name[
-                        ProjectName(requirement)
+                        requirement.project_name
                     ]
                     if contains:
                         rendered_contains = (
@@ -565,7 +685,7 @@ class PEXEnvironment(object):
                     else:
                         rendered_contains = (
                             "\n    But this pex had no {project_name!r} distributions.".format(
-                                project_name=requirement.project_name
+                                project_name=requirement.project_name.raw
                             )
                         )
                     items.append(
@@ -583,25 +703,25 @@ class PEXEnvironment(object):
                     "Failed to resolve requirements from PEX environment @ {pex}.\n"
                     "Needed {platform} compatible dependencies for:\n"
                     "{items}".format(
-                        pex=self._pex, platform=self._target.platform.tag, items="\n".join(items)
+                        pex=self.source_pex,
+                        platform=self._target.platform.tag,
+                        items="\n".join(items),
                     )
                 )
 
         return OrderedSet(resolved_dists_by_key.values())
 
-    _NAMESPACE_PACKAGE_METADATA_RESOURCE = "namespace_packages.txt"
-
     @classmethod
     def _get_namespace_packages(cls, dist):
-        if dist.has_metadata(cls._NAMESPACE_PACKAGE_METADATA_RESOURCE):
-            return list(dist.get_metadata_lines(cls._NAMESPACE_PACKAGE_METADATA_RESOURCE))
-        else:
-            return []
+        # type: (Distribution) -> Tuple[Text, ...]
+        return tuple(dist.iter_metadata_lines("namespace_packages.txt"))
 
     @classmethod
     def _declare_namespace_packages(cls, resolved_dists):
         # type: (Iterable[Distribution]) -> None
-        namespace_packages_by_dist = OrderedDict()
+        namespace_packages_by_dist = (
+            OrderedDict()
+        )  # type: OrderedDict[Distribution, Tuple[Text, ...]]
         for dist in resolved_dists:
             namespace_packages = cls._get_namespace_packages(dist)
             # NB: Dists can explicitly declare empty namespace packages lists to indicate they have none.
@@ -616,26 +736,26 @@ class PEXEnvironment(object):
         # will be active in the pex environment at runtime and, as such, care must be taken.
         #
         # Properly behaved distributions will declare a dependency on `setuptools`, in which case we
-        # use that (non-vendored) distribution. A side-effect of importing `pkg_resources` from that
+        # use that (non-vendored) distribution. A side effect of importing `pkg_resources` from that
         # distribution is that a global `pkg_resources.working_set` will be populated. For various
         # `pkg_resources` distribution discovery functions to work, that global
-        # `pkg_resources.working_set` must be built with the `sys.path` fully settled. Since all dists
-        # in the dependency set (`resolved_dists`) have already been resolved and added to the
+        # `pkg_resources.working_set` must be built with the `sys.path` fully settled. Since all
+        # dists in the dependency set (`resolved_dists`) have already been resolved and added to the
         # `sys.path` we're safe to proceed here.
         #
         # Other distributions (notably `twitter.common.*`) in the wild declare `setuptools`-specific
-        # `namespace_packages` but do not properly declare a dependency on `setuptools` which they must
-        # use to:
+        # `namespace_packages` but do not properly declare a dependency on `setuptools` which they
+        # must use to:
         # 1. Declare `namespace_packages` metadata which we just verified they have with the check
         #    above.
         # 2. Declare namespace packages at runtime via the canonical:
         #    `__import__('pkg_resources').declare_namespace(__name__)`
         #
         # For such distributions we fall back to our vendored version of `setuptools`. This is safe,
-        # since we'll only introduce our shaded version when no other standard version is present and
-        # even then tear it all down when we hand off from the bootstrap to user code.
+        # since we'll only introduce our shaded version when no other standard version is present
+        # and even then tear it all down when we hand off from the bootstrap to user code.
         pkg_resources, vendored = _import_pkg_resources()
-        if vendored:
+        if not pkg_resources or vendored:
             dists = "\n".join(
                 "\n{index}. {dist} namespace packages:\n  {ns_packages}".format(
                     index=index + 1,
@@ -644,6 +764,21 @@ class PEXEnvironment(object):
                 )
                 for index, (dist, ns_packages) in enumerate(namespace_packages_by_dist.items())
             )
+            if not pkg_resources:
+                current_interpreter = PythonInterpreter.get()
+                pex_warnings.warn(
+                    "The legacy `pkg_resources` package cannot be imported by the "
+                    "{interpreter} {version} interpreter at {path}.\n"
+                    "The following distributions need `pkg_resources` to load some legacy "
+                    "namespace packages and may fail to work properly:\n{dists}".format(
+                        interpreter=current_interpreter.identity.interpreter,
+                        version=current_interpreter.python,
+                        path=current_interpreter.binary,
+                        dists=dists,
+                    )
+                )
+                return
+
             pex_warnings.warn(
                 "The `pkg_resources` package was loaded from a pex vendored version when "
                 "declaring namespace packages defined by:\n{dists}\n\nThese distributions "
@@ -663,6 +798,10 @@ class PEXEnvironment(object):
 
         resolved = self.resolve()
         for dist in resolved:
+            # N.B.: Since there can be more than one PEXEnvironment on the PEX_PATH we take care to
+            # avoid re-installing duplicate distributions we have in common with them.
+            if dist.location in sys.path:
+                continue
             with TRACER.timed("Activating %s" % dist, V=2):
                 if self._pex_info.inherit_path == InheritPath.FALLBACK:
                     # Prepend location to sys.path.
