@@ -1,4 +1,4 @@
-# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2014 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 from __future__ import absolute_import
@@ -7,21 +7,31 @@ import json
 import os
 import zipfile
 
-from pex import layout, pex_warnings, variables
-from pex.common import can_write_dir, open_zip, safe_mkdtemp
+from pex import layout, pex_root, pex_warnings, variables
+from pex.cache import root as cache_root
+from pex.common import open_zip
 from pex.compatibility import PY2
 from pex.compatibility import string as compatibility_string
 from pex.inherit_path import InheritPath
 from pex.orderedset import OrderedSet
+from pex.os import WINDOWS
 from pex.typing import TYPE_CHECKING, cast
 from pex.variables import ENV, Variables
 from pex.venv.bin_path import BinPath
 from pex.version import __version__ as pex_version
 
 if TYPE_CHECKING:
-    from typing import Any, Dict, Mapping, Optional, Text, Union
+    # MyPy run for 2.7 does not recognize the Collection type
+    from typing import Collection  # type: ignore[attr-defined]
+    from typing import Any, Dict, Iterable, Mapping, Optional, Text, Tuple, Union
 
+    from pex.cache.dirs import VenvDir
+    from pex.dist_metadata import Requirement
+
+    # N.B.: These are expensive imports and PexInfo is used during PEX bootstrapping which we want
+    # to be as fast as possible.
     from pex.interpreter import PythonInterpreter
+    from pex.interpreter_constraints import InterpreterConstraints
 
 
 # TODO(wickman) Split this into a PexInfoBuilder/PexInfo to ensure immutability.
@@ -38,7 +48,7 @@ class PexInfo(object):
     requirements: list                 # list of requirements for this environment
 
     # Environment options
-    pex_root: string                    # root of all pex-related files eg: ~/.pex
+    pex_root: string                    # root of all pex-related files eg: ~/.cache/pex
     entry_point: string                 # entry point into this pex
     script: string                      # script to execute in this pex environment
                                         # at most one of script/entry_point can be specified
@@ -52,8 +62,6 @@ class PexInfo(object):
     """
 
     PATH = layout.PEX_INFO_PATH
-    BOOTSTRAP_CACHE = "bootstraps"
-    INSTALL_CACHE = "installed_wheels"
 
     @classmethod
     def make_build_properties(cls):
@@ -62,14 +70,9 @@ class PexInfo(object):
         }
 
     @classmethod
-    def default(cls, interpreter=None):
-        # type: (Optional[PythonInterpreter]) -> PexInfo
-        pex_info = {
-            "requirements": [],
-            "distributions": {},
-            "build_properties": cls.make_build_properties(),
-        }
-        return cls(info=pex_info)
+    def default(cls):
+        # type: () -> PexInfo
+        return cls(info={"build_properties": cls.make_build_properties()})
 
     @classmethod
     def from_pex(cls, pex):
@@ -78,10 +81,10 @@ class PexInfo(object):
             with open_zip(pex) as zf:
                 pex_info = zf.read(cls.PATH)
         elif os.path.isfile(pex):  # Venv PEX
-            with open(os.path.join(os.path.dirname(pex), cls.PATH)) as fp:
+            with open(os.path.join(os.path.dirname(pex), cls.PATH), "rb") as fp:
                 pex_info = fp.read()
         else:  # Directory (Either loose or installed) PEX
-            with open(os.path.join(pex, cls.PATH)) as fp:
+            with open(os.path.join(pex, cls.PATH), "rb") as fp:
                 pex_info = fp.read()
         return cls.from_json(pex_info)
 
@@ -105,9 +108,10 @@ class PexInfo(object):
             "venv": Variables.PEX_VENV.strip_default(env),
             "inherit_path": inherit_path,
             "ignore_errors": Variables.PEX_IGNORE_ERRORS.strip_default(env),
+            "max_install_jobs": Variables.PEX_MAX_INSTALL_JOBS.strip_default(env),
         }
         # Filter out empty entries not explicitly set in the environment.
-        return cls(info=dict((k, v) for (k, v) in pex_info.items() if v is not None))
+        return cls(info={k: v for k, v in pex_info.items() if v is not None})
 
     @classmethod
     def _parse_requirement_tuple(cls, requirement_tuple):
@@ -136,18 +140,27 @@ class PexInfo(object):
             )
         self._pex_info = dict(info) if info else {}  # type: Dict[str, Any]
         self._distributions = self._pex_info.get("distributions", {})  # type: Dict[str, str]
-        # cast as set because pex info from json must store interpreter_constraints as a list
-        self._interpreter_constraints = set(self._pex_info.get("interpreter_constraints", set()))
+
+        # N.B.: InterpreterConstraints is an expensive type to import; so we do so lazily since
+        # PexInfo is used during PEX bootstrapping which we want to be fast.
+        self._interpreter_constraints = None  # type: Optional[InterpreterConstraints]
+
         requirements = self._pex_info.get("requirements", [])
         if not isinstance(requirements, (list, tuple)):
             raise ValueError("Expected requirements to be a list, got %s" % type(requirements))
         self._requirements = OrderedSet(self._parse_requirement_tuple(req) for req in requirements)
 
-    def _get_safe(self, key):
+        self._excluded = OrderedSet(self._pex_info.get("excluded", ()))  # type: OrderedSet[str]
+        self._overridden = OrderedSet(self._pex_info.get("overridden", ()))  # type: OrderedSet[str]
+
+    def _get_python2_import_system_safe_str(self, key):
+        # type: (str) -> Optional[str]
         if key not in self._pex_info:
             return None
         value = self._pex_info[key]
-        return value.encode("utf-8") if PY2 else value
+        if PY2 and isinstance(value, unicode):
+            return str(value.encode("utf-8"))
+        return cast(str, value)
 
     @property
     def build_properties(self):
@@ -164,6 +177,36 @@ class PexInfo(object):
         self._pex_info["build_properties"] = self.make_build_properties()
         self._pex_info["build_properties"].update(value)
 
+    @property
+    def inject_python_args(self):
+        # type: () -> Tuple[str, ...]
+        return tuple(self._pex_info.get("inject_python_args", ()))
+
+    @inject_python_args.setter
+    def inject_python_args(self, value):
+        # type: (Iterable[str]) -> None
+        self._pex_info["inject_python_args"] = tuple(value)
+
+    @property
+    def inject_env(self):
+        # type: () -> Dict[str, str]
+        return dict(self._pex_info.get("inject_env", {}))
+
+    @inject_env.setter
+    def inject_env(self, value):
+        # type: (Mapping[str, str]) -> None
+        self._pex_info["inject_env"] = dict(value)
+
+    @property
+    def inject_args(self):
+        # type: () -> Tuple[str, ...]
+        return tuple(self._pex_info.get("inject_args", ()))
+
+    @inject_args.setter
+    def inject_args(self, value):
+        # type: (Iterable[str]) -> None
+        self._pex_info["inject_args"] = tuple(value)
+
     @property
     def venv(self):
         # type: () -> bool
@@ -210,25 +253,68 @@ class PexInfo(object):
         # type: (bool) -> None
         self._pex_info["venv_site_packages_copies"] = value
 
-    def venv_dir(
+    @property
+    def venv_system_site_packages(self):
+        # type: () -> bool
+        return self._pex_info.get("venv_system_site_packages", False)
+
+    @venv_system_site_packages.setter
+    def venv_system_site_packages(self, value):
+        # type: (bool) -> None
+        self._pex_info["venv_system_site_packages"] = value
+
+    @property
+    def venv_hermetic_scripts(self):
+        # type: () -> bool
+        return self._pex_info.get("venv_hermetic_scripts", True)
+
+    @venv_hermetic_scripts.setter
+    def venv_hermetic_scripts(self, value):
+        # type: (bool) -> None
+        self._pex_info["venv_hermetic_scripts"] = value
+
+    def _venv_dir(
         self,
+        pex_root,  # type: str
         pex_file,  # type: str
         interpreter=None,  # type: Optional[PythonInterpreter]
+        expand_pex_root=True,  # type: bool
     ):
-        # type: (...) -> Optional[str]
+        # type: (...) -> Optional[VenvDir]
         if not self.venv:
             return None
         if self.pex_hash is None:
             raise ValueError("The venv_dir was requested but no pex_hash was set.")
         return variables.venv_dir(
-            pex_file=pex_file,
-            pex_root=self.pex_root,
+            pex_root=pex_root,
             pex_hash=self.pex_hash,
-            has_interpreter_constraints=bool(self.interpreter_constraints),
+            has_interpreter_constraints=self.has_interpreter_constraints,
+            pex_file=pex_file,
             interpreter=interpreter,
             pex_path=self.pex_path,
+            expand_pex_root=expand_pex_root,
         )
 
+    def runtime_venv_dir(
+        self,
+        pex_file,  # type: str
+        interpreter=None,  # type: Optional[PythonInterpreter]
+    ):
+        # type: (...) -> Optional[VenvDir]
+        return self._venv_dir(self.pex_root, pex_file, interpreter)
+
+    def raw_venv_dir(
+        self,
+        pex_file,  # type: str
+        interpreter=None,  # type: Optional[PythonInterpreter]
+    ):
+        # type: (...) -> Optional[VenvDir]
+        """Distiguished from ``venv_dir`` by use of the raw_pex_root.
+        We don't expand the pex_root at build time in case the pex_root is not
+        writable or doesn't exist at build time.
+        """
+        return self._venv_dir(self.raw_pex_root, pex_file, interpreter, expand_pex_root=False)
+
     @property
     def includes_tools(self):
         # type: () -> bool
@@ -254,18 +340,31 @@ class PexInfo(object):
 
     @property
     def pex_path(self):
-        # type: () -> Optional[str]
-        """A colon separated list of other pex files to merge into the runtime environment.
+        # type: () -> Tuple[str, ...]
+        """A list of other pex files to merge into the runtime environment.
 
         This pex info property is used to persist the PEX_PATH environment variable into the pex
         info metadata for reuse within a built pex.
         """
-        return cast("Optional[str]", self._pex_info.get("pex_path"))
+        pex_paths = self._pex_info.get("pex_paths")
+        if pex_paths:
+            return tuple(cast("Iterable[str]", pex_paths))
+
+        # Legacy PEX-INFO stored this in a single string as a colon-separated list.
+        pex_path = self._pex_info.get("pex_path")
+        if pex_path:
+            return tuple(pex_path.split(":"))
+
+        return ()
 
     @pex_path.setter
     def pex_path(self, value):
-        # type: (str) -> None
-        self._pex_info["pex_path"] = value
+        # type: (Iterable[str]) -> None
+        if not WINDOWS:
+            # Store in the legacy format on the legacy supported OSes for backwards compatibility
+            # with old tools reading new PEX-INFO.
+            self._pex_info["pex_path"] = ":".join(value)
+        self._pex_info["pex_paths"] = tuple(value)
 
     @property
     def inherit_path(self):
@@ -286,8 +385,16 @@ class PexInfo(object):
         # type: (InheritPath.Value) -> None
         self._pex_info["inherit_path"] = value.value
 
+    @property
+    def has_interpreter_constraints(self):
+        # type: () -> bool
+        if self._interpreter_constraints is not None:
+            return bool(self._interpreter_constraints)
+        return bool(self._pex_info.get("interpreter_constraints", ()))
+
     @property
     def interpreter_constraints(self):
+        # type: () -> InterpreterConstraints
         """A list of constraints that determine the interpreter compatibility for this pex, using
         the Requirement-style format, e.g. ``'CPython>=3', or just '>=2.7,<3'`` for requirements
         agnostic to interpreter class.
@@ -295,10 +402,21 @@ class PexInfo(object):
         This property will be used at exec time when bootstrapping a pex to search PEX_PYTHON_PATH
         for a list of compatible interpreters.
         """
-        return list(self._interpreter_constraints)
+        if self._interpreter_constraints is None:
+            # N.B.: InterpreterConstraints is an expensive import and PexInfo is used during PEX
+            # bootstrapping which we want to be as fast as possible; so we go through hoops to be
+            # lazy here.
+            from pex.interpreter_constraints import InterpreterConstraints
 
-    def add_interpreter_constraint(self, value):
-        self._interpreter_constraints.add(str(value))
+            self._interpreter_constraints = InterpreterConstraints.parse(
+                *self._pex_info.get("interpreter_constraints", ())
+            )
+        return self._interpreter_constraints
+
+    @interpreter_constraints.setter
+    def interpreter_constraints(self, value):
+        # type: (InterpreterConstraints) -> None
+        self._interpreter_constraints = value
 
     @property
     def ignore_errors(self):
@@ -338,18 +456,22 @@ class PexInfo(object):
 
     @property
     def entry_point(self):
-        return self._get_safe("entry_point")
+        # type: () -> Optional[str]
+        return self._get_python2_import_system_safe_str("entry_point")
 
     @entry_point.setter
     def entry_point(self, value):
+        # type: (str) -> None
         self._pex_info["entry_point"] = value
 
     @property
     def script(self):
-        return self._get_safe("script")
+        # type: () -> Optional[str]
+        return self._get_python2_import_system_safe_str("script")
 
     @script.setter
     def script(self, value):
+        # type: (str) -> None
         self._pex_info["script"] = value
 
     def add_requirement(self, requirement):
@@ -357,8 +479,27 @@ class PexInfo(object):
 
     @property
     def requirements(self):
+        # type: () -> Collection[str]
         return self._requirements
 
+    def add_exclude(self, requirement):
+        # type: (Requirement) -> None
+        self._excluded.add(str(requirement))
+
+    @property
+    def excluded(self):
+        # type: () -> Iterable[str]
+        return self._excluded
+
+    def add_override(self, requirement):
+        # type: (Requirement) -> None
+        self._overridden.add(str(requirement))
+
+    @property
+    def overridden(self):
+        # type: () -> Iterable[str]
+        return self._overridden
+
     def add_distribution(self, location, sha):
         self._distributions[location] = sha
 
@@ -370,21 +511,15 @@ class PexInfo(object):
     @property
     def raw_pex_root(self):
         # type: () -> str
-        return cast(str, self._pex_info.get("pex_root", os.path.join("~", ".pex")))
+        return cast(str, self._pex_info.get("pex_root", cache_root.path(expand_user=False)))
 
     @property
     def pex_root(self):
         # type: () -> str
-        pex_root = os.path.expanduser(self.raw_pex_root)
-        if not can_write_dir(pex_root):
-            tmp_root = safe_mkdtemp()
-            pex_warnings.warn(
-                "PEX_ROOT is configured as {pex_root} but that path is un-writeable, "
-                "falling back to a temporary PEX_ROOT of {tmp_root} which will hurt "
-                "performance.".format(pex_root=pex_root, tmp_root=tmp_root)
-            )
-            pex_root = self._pex_info["pex_root"] = tmp_root
-        return pex_root
+        writeable_pex_root, is_fallback = pex_root.ensure_writeable(self.raw_pex_root)
+        if is_fallback:
+            self._pex_info["pex_root"] = writeable_pex_root
+        return writeable_pex_root
 
     @pex_root.setter
     def pex_root(self, value):
@@ -404,74 +539,85 @@ class PexInfo(object):
         # type: (str) -> None
         self._pex_info["bootstrap_hash"] = value
 
+    @property
+    def deps_are_wheel_files(self):
+        # type: () -> bool
+        return self._pex_info.get("deps_are_wheel_files", False)
+
+    @deps_are_wheel_files.setter
+    def deps_are_wheel_files(self, value):
+        # type: (bool) -> None
+        self._pex_info["deps_are_wheel_files"] = value
+
+    @property
+    def max_install_jobs(self):
+        # type: () -> int
+        return self._pex_info.get("max_install_jobs", 1)
+
+    @max_install_jobs.setter
+    def max_install_jobs(self, value):
+        # type: (int) -> None
+        if value < -1:
+            raise ValueError(
+                "The value for max_install_jobs must be -1 or greater; given: {jobs}".format(
+                    jobs=value
+                )
+            )
+        self._pex_info["max_install_jobs"] = value
+
     @property
     def bootstrap(self):
         # type: () -> str
         return layout.BOOTSTRAP_DIR
 
-    @property
-    def bootstrap_cache(self):
-        # type: () -> Optional[str]
-        if self.bootstrap_hash is None:
-            return None
-        return os.path.join(self.pex_root, self.BOOTSTRAP_CACHE, self.bootstrap_hash)
-
     @property
     def internal_cache(self):
         # type: () -> str
         return layout.DEPS_DIR
 
-    @property
-    def install_cache(self):
-        return os.path.join(self.pex_root, self.INSTALL_CACHE)
-
-    @property
-    def zip_unsafe_cache(self):
-        #: type: () -> str
-        return os.path.join(self.pex_root, "user_code")
-
     def update(self, other):
         # type: (PexInfo) -> None
         if not isinstance(other, PexInfo):
             raise TypeError("Cannot merge a %r with PexInfo" % type(other))
         self._pex_info.update(other._pex_info)
         self._distributions.update(other.distributions)
-        self._interpreter_constraints.update(other.interpreter_constraints)
+        if other.has_interpreter_constraints:
+            self._interpreter_constraints = self.interpreter_constraints.merged(
+                other.interpreter_constraints
+            )
         self._requirements.update(other.requirements)
+        self._excluded.update(other.excluded)
 
     def as_json_dict(self):
         # type: () -> Dict[str, Any]
         data = self._pex_info.copy()
         data["inherit_path"] = self.inherit_path.value
-        data["requirements"] = list(self._requirements)
-        data["interpreter_constraints"] = list(self._interpreter_constraints)
+        data["requirements"] = sorted(self._requirements)
+        data["excluded"] = sorted(self._excluded)
+        data["overridden"] = sorted(self._overridden)
+        data["interpreter_constraints"] = sorted(str(ic) for ic in self.interpreter_constraints)
         data["distributions"] = self._distributions.copy()
         return data
 
-    def dump(self):
-        # type: (...) -> str
-        data = self.as_json_dict()
-        data["requirements"].sort()
-        data["interpreter_constraints"].sort()
-        return json.dumps(data, sort_keys=True)
+    def dump(self, **extra_json_dumps_kwargs):
+        # type: (**Any) -> str
+        return json.dumps(self.as_json_dict(), sort_keys=True, **extra_json_dumps_kwargs)
 
     def copy(self):
         # type: () -> PexInfo
         return PexInfo(self.as_json_dict())
 
-    @staticmethod
-    def _merge_split(*paths):
-        filtered_paths = filter(None, paths)
-        return [p for p in ":".join(filtered_paths).split(":") if p]
-
     def merge_pex_path(self, pex_path):
+        # type: (Iterable[str]) -> None
         """Merges a new PEX_PATH definition into the existing one (if any).
 
-        :param str pex_path: The PEX_PATH to merge.
+        :param pex_path: The PEX_PATH to merge.
         """
         if not pex_path:
             return
-        self.pex_path = ":".join(self._merge_split(self.pex_path, pex_path))
+        merged_pex_path = OrderedSet(self.pex_path)
+        merged_pex_path.update(pex_path)
+        self.pex_path = tuple(merged_pex_path)
 
     def __repr__(self):
         return "{}({!r})".format(type(self).__name__, self._pex_info)
