@@ -1,4 +1,4 @@
-# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2015 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 from __future__ import absolute_import
@@ -7,19 +7,24 @@ import os
 import re
 from contextlib import contextmanager
 
-from pex import attrs, dist_metadata
-from pex.compatibility import urlparse
-from pex.dist_metadata import MetadataError, ProjectNameAndVersion
-from pex.enum import Enum
+from pex import attrs, dist_metadata, pex_warnings
+from pex.artifact_url import VCS, ArchiveScheme, ArtifactURL, VCSScheme
+from pex.compatibility import url_unquote, urlparse
+from pex.dist_metadata import (
+    MetadataError,
+    ProjectNameAndVersion,
+    Requirement,
+    RequirementParseError,
+)
 from pex.fetcher import URLFetcher
+from pex.pep_503 import ProjectName
 from pex.third_party.packaging.markers import Marker
 from pex.third_party.packaging.specifiers import SpecifierSet
 from pex.third_party.packaging.version import InvalidVersion, Version
-from pex.third_party.pkg_resources import Requirement, RequirementParseError
-from pex.typing import TYPE_CHECKING, cast
+from pex.typing import TYPE_CHECKING
 
 if TYPE_CHECKING:
-    from typing import Iterable, Iterator, Match, Optional, Text, Tuple, Union
+    from typing import FrozenSet, Iterable, Iterator, Match, Optional, Text, Tuple, Union
 
     import attr  # vendor:skip
 else:
@@ -52,6 +57,11 @@ class Source(object):
         is_constraints=False,  # type: bool
     ):
         # type: (...) -> Iterator[Source]
+        pex_warnings.warn(
+            "Fetching {subject} files via url is deprecated: {url}".format(
+                url=url, subject="constraints" if is_constraints else "requirements"
+            )
+        )
         with fetcher.get_body_iter(url) as lines:
             yield cls(origin=url, is_file=False, is_constraints=is_constraints, lines=lines)
 
@@ -119,6 +129,8 @@ class Source(object):
             return
 
         path = url.path if url.scheme == "file" else origin
+        if not os.path.isabs(path) and self.is_file:
+            path = os.path.join(os.path.dirname(self.origin), path)
         try:
             with self.from_file(path, is_constraints=is_constraints) as source:
                 yield source
@@ -127,43 +139,66 @@ class Source(object):
 
 
 @attr.s(frozen=True)
-class PyPIRequirement(object):
+class _ParsedRequirement(object):
+    line = attr.ib()  # type: LogicalLine
+
+    def __str__(self):
+        # type: () -> str
+        return str(self.line.processed_text)
+
+
+@attr.s(frozen=True)
+class PyPIRequirement(_ParsedRequirement):
     """A requirement realized through a package index or find links repository."""
 
-    line = attr.ib()  # type: LogicalLine
     requirement = attr.ib()  # type: Requirement
-    editable = attr.ib(default=False)  # type: bool
+
+    @property
+    def extras(self):
+        # type: () -> FrozenSet[str]
+        return self.requirement.extras
 
 
 @attr.s(frozen=True)
-class URLRequirement(object):
+class URLRequirement(_ParsedRequirement):
     """A requirement realized through a distribution archive at a fixed URL."""
 
-    line = attr.ib()  # type: LogicalLine
-    url = attr.ib()  # type: Text
+    url = attr.ib()  # type: ArtifactURL
     requirement = attr.ib()  # type: Requirement
-    editable = attr.ib(default=False)  # type: bool
 
+    @property
+    def extras(self):
+        # type: () -> FrozenSet[str]
+        return self.requirement.extras
 
-class VCS(Enum["VCS.Value"]):
-    class Value(Enum.Value):
-        pass
+    @property
+    def filename(self):
+        # type: () -> str
+        return os.path.basename(self.url.path)
 
-    Bazaar = Value("bzr")
-    Git = Value("git")
-    Mercurial = Value("hg")
-    Subversion = Value("svn")
+    @property
+    def project_name(self):
+        return self.requirement.project_name
+
+    @property
+    def subdirectory(self):
+        # type: () -> Optional[str]
+        subdirectories = self.url.fragment_parameters.get("subdirectory")
+        return subdirectories[-1] if subdirectories else None
 
 
 @attr.s(frozen=True)
-class VCSRequirement(object):
+class VCSRequirement(_ParsedRequirement):
     """A requirement realized by building a distribution from sources retrieved from a VCS."""
 
-    line = attr.ib()  # type: LogicalLine
     vcs = attr.ib()  # type: VCS.Value
     url = attr.ib()  # type: Text
     requirement = attr.ib()  # type: Requirement
-    editable = attr.ib(default=False)  # type: bool
+
+    @property
+    def extras(self):
+        # type: () -> FrozenSet[str]
+        return self.requirement.extras
 
 
 def parse_requirement_from_project_name_and_specifier(
@@ -171,6 +206,8 @@ def parse_requirement_from_project_name_and_specifier(
     extras=None,  # type: Optional[Iterable[str]]
     specifier=None,  # type: Optional[SpecifierSet]
     marker=None,  # type: Optional[Marker]
+    editable=False,  # type: bool
+    url=None,  # type: Optional[Text]
 ):
     # type: (...) -> Requirement
     requirement_string = "{project_name}{extras}{specifier}".format(
@@ -180,13 +217,14 @@ def parse_requirement_from_project_name_and_specifier(
     )
     if marker:
         requirement_string += ";" + str(marker)
-    return Requirement.parse(requirement_string)
+    return attr.evolve(Requirement.parse(requirement_string, editable=editable), url=url)
 
 
 def parse_requirement_from_dist(
     dist,  # type: str
     extras=None,  # type: Optional[Iterable[str]]
     marker=None,  # type: Optional[Marker]
+    editable=False,  # type: bool
 ):
     # type: (...) -> Requirement
     project_name_and_version = dist_metadata.project_name_and_version(dist)
@@ -203,23 +241,39 @@ def parse_requirement_from_dist(
         extras=extras,
         specifier=project_name_and_specifier.specifier,
         marker=marker,
+        editable=editable,
     )
 
 
 @attr.s(frozen=True)
-class LocalProjectRequirement(object):
+class LocalProjectRequirement(_ParsedRequirement):
     """A requirement realized by building a distribution from local sources."""
 
-    line = attr.ib()  # type: LogicalLine
     path = attr.ib()  # type: str
     extras = attr.ib(default=(), converter=attrs.str_tuple_from_iterable)  # type: Tuple[str, ...]
     marker = attr.ib(default=None)  # type: Optional[Marker]
     editable = attr.ib(default=False)  # type: bool
+    project_name = attr.ib(default=None)  # type: Optional[ProjectName]
 
-    def as_requirement(self, dist):
-        # type: (str) -> Requirement
+    def as_requirement(self, dist=None):
+        # type: (Optional[str]) -> Requirement
         """Create a requirement given a distribution that was built from this local project."""
-        return parse_requirement_from_dist(dist, self.extras, self.marker)
+        if dist is None and self.project_name is None:
+            raise ValueError(
+                "No distribution was supplied and the local project at {path} has an unknown "
+                "project name; so no requirement can be calculated.".format(path=self.path)
+            )
+
+        if dist is not None:
+            return parse_requirement_from_dist(dist, self.extras, self.marker, self.editable)
+
+        return parse_requirement_from_project_name_and_specifier(
+            project_name=str(self.project_name),
+            extras=self.extras,
+            marker=self.marker,
+            editable=self.editable,
+            url="file://{path}".format(path=self.path),
+        )
 
 
 if TYPE_CHECKING:
@@ -229,8 +283,7 @@ if TYPE_CHECKING:
 
 
 @attr.s(frozen=True)
-class Constraint(object):
-    line = attr.ib()  # type: LogicalLine
+class Constraint(_ParsedRequirement):
     requirement = attr.ib()  # type: Requirement
 
 
@@ -260,56 +313,6 @@ def _strip_requirement_options(line):
     return editable, re.sub(r"\s--(global-option|install-option|hash).*$", "", processed_text)
 
 
-class ArchiveScheme(Enum["ArchiveScheme.Value"]):
-    class Value(Enum.Value):
-        pass
-
-    FTP = Value("ftp")
-    HTTP = Value("http")
-    HTTPS = Value("https")
-
-
-@attr.s(frozen=True)
-class VCSScheme(object):
-    vcs = attr.ib()  # type: VCS.Value
-    scheme = attr.ib()  # type: str
-
-
-def _parse_non_local_scheme(scheme):
-    # type: (str) -> Optional[Union[ArchiveScheme.Value, VCSScheme]]
-    match = re.match(
-        r"""
-        ^
-        (?:
-            (?P<archive_scheme>
-                # Archives
-                  ftp
-                | https?
-            )
-            |
-            (?P<vcs_type>
-                # VCSs: https://pip.pypa.io/en/stable/reference/pip_install/#vcs-support       
-                  bzr
-                | git
-                | hg
-                | svn
-            )\+(?P<vcs_scheme>.+)
-        )
-        $
-        """,
-        scheme,
-        re.VERBOSE,
-    )
-    if not match:
-        return None
-
-    archive_scheme = match.group("archive_scheme")
-    if archive_scheme:
-        return cast(ArchiveScheme.Value, ArchiveScheme.for_value(archive_scheme))
-
-    return VCSScheme(vcs=VCS.for_value(match.group("vcs_type")), scheme=match.group("vcs_scheme"))
-
-
 @attr.s(frozen=True)
 class ProjectNameExtrasAndMarker(object):
     project_name = attr.ib()  # type: Text
@@ -321,27 +324,25 @@ class ProjectNameExtrasAndMarker(object):
         return self.project_name, self.extras, self.marker
 
 
-def _try_parse_fragment_project_name_and_marker(fragment):
-    # type: (Text) -> Optional[ProjectNameExtrasAndMarker]
-    project_requirement = None
-    for part in fragment.split("&"):
-        if part.startswith("egg="):
-            _, project_requirement = part.split("=", 1)
-            break
-    if project_requirement is None:
+def _try_parse_fragment_project_name_and_marker(url):
+    # type: (ArtifactURL) -> Optional[ProjectNameExtrasAndMarker]
+    project_names = url.fragment_parameters.get("egg")
+    if not project_names:
         return None
+
+    project_name = project_names[-1]
     try:
-        req = Requirement.parse(project_requirement)
+        req = Requirement.parse(project_name)
         return ProjectNameExtrasAndMarker(req.name, extras=req.extras, marker=req.marker)
     except (RequirementParseError, ValueError):
-        return ProjectNameExtrasAndMarker(project_requirement)
+        return ProjectNameExtrasAndMarker(project_name)
 
 
 @attr.s(frozen=True)
 class ProjectNameAndSpecifier(object):
     @staticmethod
     def _version_as_specifier(version):
-        # type: (str) -> SpecifierSet
+        # type: (Text) -> SpecifierSet
         try:
             return SpecifierSet("=={}".format(Version(version)))
         except InvalidVersion:
@@ -379,14 +380,16 @@ def _try_parse_pip_local_formats(
     # Requirements strings can optionally include:
     REQUIREMENT_PARTS_START = (
         # + Trailing extras denoted by `[...]`.
-        #   See: https://www.python.org/dev/peps/pep-0508/#extras
+        #   See: https://peps.python.org/pep-0508/#extras
         r"\[",
         # + A version specifier denoted by a leading `!=`, `==`, `===`, `>=`, `<=` or `~=`.
-        #   See: https://www.python.org/dev/peps/pep-0508/#grammar
+        #   See: https://peps.python.org/pep-0508/#grammar
         r"!=><~",
         # + Environment markers denoted by `;...`
-        #   See: https://www.python.org/dev/peps/pep-0508/#environment-markers
+        #   See: https://peps.python.org/pep-0508/#environment-markers
         r";",
+        # + A trailing whitespace character separating any of the above
+        r"\s",
     )
     # N.B.: The basename of the current directory (.) is '' and we allow this.
     match = re.match(
@@ -407,7 +410,9 @@ def _try_parse_pip_local_formats(
     directory_name, requirement_parts = match.groups()
     stripped_path = os.path.join(os.path.dirname(path), directory_name)
     abs_stripped_path = (
-        os.path.join(basepath, stripped_path) if basepath else os.path.abspath(stripped_path)
+        os.path.join(basepath, stripped_path)
+        if basepath and not os.path.isabs(stripped_path)
+        else os.path.abspath(stripped_path)
     )
     if not os.path.exists(abs_stripped_path):
         return None
@@ -456,14 +461,11 @@ def _parse_requirement_line(
 
     editable, processed_text = _strip_requirement_options(line)
     project_name, direct_reference_url = _split_direct_references(processed_text)
-    parsed_url = urlparse.urlparse(direct_reference_url or processed_text)
+    parsed_url = ArtifactURL.parse(direct_reference_url or processed_text)
 
-    # Handle non local URLs.
-    non_local_scheme = _parse_non_local_scheme(parsed_url.scheme)
-    if non_local_scheme:
-        project_name_extras_and_marker = _try_parse_fragment_project_name_and_marker(
-            parsed_url.fragment
-        )
+    # Handle non-local URLs.
+    if isinstance(parsed_url.scheme, (ArchiveScheme.Value, VCSScheme)):
+        project_name_extras_and_marker = _try_parse_fragment_project_name_and_marker(parsed_url)
         project_name, extras, marker = (
             project_name_extras_and_marker.astuple()
             if project_name_extras_and_marker
@@ -472,7 +474,9 @@ def _parse_requirement_line(
         specifier = None  # type: Optional[SpecifierSet]
         if not project_name:
             project_name_and_specifier = _try_parse_project_name_and_specifier_from_path(
-                parsed_url.path
+                # There may be whitespace separated markers; so we strip the trailing whitespace
+                # used to support those.
+                url_unquote(parsed_url.path).rstrip()
             )
             if project_name_and_specifier is not None:
                 project_name = project_name_and_specifier.project_name
@@ -480,30 +484,41 @@ def _parse_requirement_line(
         # Pip allows an environment marker after the url which matches the urlparse structure:
         #   scheme://netloc/path;parameters?query#fragment
         # See: https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlparse
-        if not marker and parsed_url.params:
-            marker = Marker(parsed_url.params)
+        if not marker and parsed_url.parameters:
+            marker = Marker(parsed_url.parameters)
         if project_name is None:
             raise ParseError(
                 line,
                 (
-                    "Could not determine a project name for URL requirement {}, consider using "
-                    "#egg=<project name>."
+                    "Could not determine a project name for URL requirement {url}, consider using "
+                    "#egg=<project name>.".format(url=parsed_url.raw_url)
                 ),
             )
-        url = parsed_url._replace(params="", fragment="").geturl()
+        parsed_url_info = parsed_url.url_info._replace(
+            params="",
+            # Omit `egg=project_name` since that gets moved to a `project_name @` suffix in
+            # `parse_requirement_from_project_name_and_specifier`, but retain all other fragment
+            # parameters; notably, subdirectory when present.
+            fragment=parsed_url.fragment(excludes={"egg"}),
+        )
+        # There may be whitespace separated markers; so we strip the trailing whitespace
+        # used to support those.
+        url = parsed_url_info.geturl().rstrip()
         requirement = parse_requirement_from_project_name_and_specifier(
             project_name,
             extras=extras,
             specifier=specifier,
             marker=marker,
+            url=url,
         )
-        if isinstance(non_local_scheme, VCSScheme):
-            url = urlparse.urlparse(url)._replace(scheme=non_local_scheme.scheme).geturl()
-            return VCSRequirement(line, non_local_scheme.vcs, url, requirement, editable=editable)
-        return URLRequirement(line, url, requirement, editable=editable)
+        parsed_scheme = parsed_url.scheme
+        if isinstance(parsed_scheme, VCSScheme):
+            url = parsed_url_info._replace(scheme=parsed_scheme.scheme).geturl()
+            return VCSRequirement(line, parsed_scheme.vcs, url, requirement)
+        return URLRequirement(line, url=ArtifactURL.parse(url), requirement=requirement)
 
     # Handle local archives and project directories via path or file URL (Pip proprietary).
-    local_requirement = parsed_url._replace(scheme="").geturl()
+    local_requirement = parsed_url.url_info._replace(scheme="").geturl()
     project_name_extras_and_marker = _try_parse_pip_local_formats(
         local_requirement, basepath=basepath
     )
@@ -529,19 +544,23 @@ def _parse_requirement_line(
             requirement = parse_requirement_from_dist(
                 archive_or_project_path, extras=extras, marker=marker
             )
-            return URLRequirement(line, archive_or_project_path, requirement, editable=editable)
+            return URLRequirement(
+                line,
+                url=ArtifactURL.parse(archive_or_project_path),
+                requirement=requirement,
+            )
         except dist_metadata.UnrecognizedDistributionFormat:
             # This is not a recognized local archive distribution. Fall through and try parsing as a
             # PEP-440 requirement.
             pass
 
-    # Handle PEP-440. See: https://www.python.org/dev/peps/pep-0440.
+    # Handle PEP-440. See: https://peps.python.org/pep-0440.
     #
-    # The `pkg_resources.Requirement.parse` method does all of this for us (via
+    # The `pex.dist_metadata.Requirement.parse` method does all of this for us (via
     # `packaging.requirements.Requirement`) except for the handling of PEP-440 direct url
     # references; which we handled above and won't encounter here.
     try:
-        return PyPIRequirement(line, Requirement.parse(processed_text), editable=editable)
+        return PyPIRequirement(line, Requirement.parse(processed_text))
     except RequirementParseError as e:
         raise ParseError(
             line, "Problem parsing {!r} as a requirement: {}".format(processed_text, e)
@@ -635,8 +654,11 @@ def parse_requirements(
                         yield requirement
                 continue
 
-            # Skip empty lines, comment lines and all other Pip options.
-            if not processed_text or processed_text.startswith("-"):
+            # Skip empty lines, comment lines and all Pip global options.
+            if not processed_text or (
+                processed_text.startswith("-")
+                and not re.match(r"^(?:-e|--editable)\s.*", processed_text)
+            ):
                 continue
 
             # Only requirement lines remain.
@@ -686,15 +708,20 @@ def parse_requirement_file(
             yield req_info
 
 
+def parse_requirement_string(requirement):
+    # type: (Text) -> ParsedRequirement
+    return _parse_requirement_line(
+        LogicalLine(
+            raw_text=requirement,
+            processed_text=requirement.strip(),
+            source="<string>",
+            start_line=1,
+            end_line=1,
+        )
+    )
+
+
 def parse_requirement_strings(requirements):
     # type: (Iterable[Text]) -> Iterator[ParsedRequirement]
     for requirement in requirements:
-        yield _parse_requirement_line(
-            LogicalLine(
-                raw_text=requirement,
-                processed_text=requirement.strip(),
-                source="<string>",
-                start_line=1,
-                end_line=1,
-            )
-        )
+        yield parse_requirement_string(requirement)
