@@ -1,11 +1,10 @@
-# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2021 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 import os
-import subprocess
 
 from pex.common import temporary_dir
-from pex.testing import make_env, run_pex_command
+from testing import make_env, run_pex_command, subprocess
 
 
 def test_force_local_implicit_ns_packages():
