@@ -1,36 +1,37 @@
-# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2021 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 import json
 import os
 import re
-import subprocess
 from textwrap import dedent
 
 import pytest
 
-import pex.resolve.lockfile
+from pex.artifact_url import ArtifactURL, Fingerprint
 from pex.compatibility import PY2
+from pex.dist_metadata import Constraint, Requirement
 from pex.pep_440 import Version
 from pex.pep_503 import ProjectName
+from pex.pip.version import PipVersion
 from pex.resolve.locked_resolve import Artifact, LockedRequirement, LockedResolve, LockStyle
-from pex.resolve.lockfile import Lockfile, json_codec
-from pex.resolve.resolved_requirement import Fingerprint, Pin
-from pex.resolve.resolver_configuration import ResolverVersion
+from pex.resolve.lockfile import json_codec
+from pex.resolve.lockfile.json_codec import ParseError, PathMappingError
+from pex.resolve.lockfile.model import Lockfile
+from pex.resolve.path_mappings import PathMapping, PathMappings
+from pex.resolve.resolved_requirement import Pin
+from pex.resolve.resolver_configuration import BuildConfiguration, ResolverVersion
 from pex.sorted_tuple import SortedTuple
 from pex.third_party.packaging import tags
-from pex.third_party.pkg_resources import Requirement
 from pex.typing import TYPE_CHECKING
+from testing import subprocess
 
 if TYPE_CHECKING:
-    from typing import Any
+    from typing import Any, Container
 
     import attr  # vendor:skip
 else:
-    if "__PEX_UNVENDORED__" in __import__("os").environ:
-        import attr  # vendor:skip
-    else:
-        import pex.third_party.attr as attr
+    import pex.third_party.attr as attr
 
 
 def test_roundtrip(tmpdir):
@@ -40,19 +41,19 @@ def test_roundtrip(tmpdir):
         pex_version="1.2.3",
         style=LockStyle.STRICT,
         requires_python=(),
+        target_systems=(),
+        pip_version=PipVersion.VENDORED,
         resolver_version=ResolverVersion.PIP_2020,
         requirements=(
             Requirement.parse("ansicolors"),
             Requirement.parse("requests>=2; sys_platform == 'darwin'"),
         ),
-        constraints=(Requirement.parse("ansicolors==1.1.8"),),
+        constraints=(Constraint.parse("ansicolors==1.1.8"),),
         allow_prereleases=True,
-        allow_wheels=False,
-        allow_builds=False,
-        prefer_older_binary=False,
-        use_pep517=None,
-        build_isolation=True,
+        build_configuration=BuildConfiguration(),
         transitive=False,
+        excluded=(),
+        overridden=(),
         locked_resolves=[
             LockedResolve(
                 platform_tag=tags.Tag("cp36", "cp36m", "macosx_10_13_x86_64"),
@@ -62,7 +63,7 @@ def test_roundtrip(tmpdir):
                             pin=Pin(
                                 project_name=ProjectName("ansicolors"), version=Version("1.1.8")
                             ),
-                            artifact=Artifact(
+                            artifact=Artifact.from_url(
                                 url="https://example.org/colors-1.1.8-cp36-cp36m-macosx_10_6_x86_64.whl",
                                 fingerprint=Fingerprint(algorithm="blake256", hash="cafebabe"),
                             ),
@@ -70,12 +71,12 @@ def test_roundtrip(tmpdir):
                         ),
                         LockedRequirement.create(
                             pin=Pin(project_name=ProjectName("requests"), version=Version("2.0.0")),
-                            artifact=Artifact(
+                            artifact=Artifact.from_url(
                                 url="https://example.org/requests-2.0.0-py2.py3-none-any.whl",
                                 fingerprint=Fingerprint(algorithm="sha256", hash="456"),
                             ),
                             additional_artifacts=(
-                                Artifact(
+                                Artifact.from_url(
                                     url="file://find-links/requests-2.0.0.tar.gz",
                                     fingerprint=Fingerprint(algorithm="sha512", hash="123"),
                                 ),
@@ -92,7 +93,7 @@ def test_roundtrip(tmpdir):
                             pin=Pin(
                                 project_name=ProjectName("ansicolors"), version=Version("1.1.8")
                             ),
-                            artifact=Artifact(
+                            artifact=Artifact.from_url(
                                 url="https://example.org/colors-1.1.8-cp37-cp37m-manylinux1_x86_64.whl",
                                 fingerprint=Fingerprint(algorithm="md5", hash="hackme"),
                             ),
@@ -125,7 +126,7 @@ VALID_LOCK = """\
             {
               "algorithm": "md5",
               "hash": "f357aa02db2466bc24ff1815cff1aeb3",
-              "url": "http://localhost:9999/ansicolors-1.1.8-py2.py3-none-any.whl"
+              "url": "file:///find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
             },
             {
               "algorithm": "md5",
@@ -191,13 +192,24 @@ def patch_tool(tmpdir):
     return PatchTool(str(tmpdir))
 
 
+def parse_patched(
+    patch_tool,  # type: PatchTool
+    patch,  # type: str
+    path_mappings=PathMappings(),  # type: PathMappings
+):
+    # type: (...) -> Lockfile
+    return json_codec.loads(lockfile_contents=patch_tool.apply(patch), path_mappings=path_mappings)
+
+
 def assert_parse_error(
     patch_tool,  # type: PatchTool
     patch,  # type: str
     match,  # type: str
+    path_mappings=PathMappings(),  # type: PathMappings
 ):
-    with pytest.raises(pex.resolve.lockfile.ParseError, match=match):
-        json_codec.loads(patch_tool.apply(patch))
+    # type: (...) -> None
+    with pytest.raises(ParseError, match=match):
+        parse_patched(patch_tool, patch, path_mappings=path_mappings)
 
 
 def test_load_invalid_json(patch_tool):
@@ -264,7 +276,7 @@ def test_load_invalid_key_not_found(patch_tool):
                            "algorithm": "md5",
             -              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
             +              "HASH": "f357aa02db2466bc24ff1815cff1aeb3",
-                           "url": "http://localhost:9999/ansicolors-1.1.8-py2.py3-none-any.whl"
+                           "url": "file:///find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
             """
         ),
         match=re.escape(
@@ -292,7 +304,7 @@ def test_load_invalid_parent_not_json_object(patch_tool):
             -            {
             -              "algorithm": "md5",
             -              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
-            -              "url": "http://localhost:9999/ansicolors-1.1.8-py2.py3-none-any.whl"
+            -              "url": "file:///find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
             -            },
             +            "foo",
                          {
@@ -455,23 +467,6 @@ def test_load_invalid_platform_tag(patch_tool):
     )
 
 
-def test_load_invalid_no_requirements(patch_tool):
-    # type: (PatchTool) -> None
-
-    assert_parse_error(
-        patch_tool,
-        dedent(
-            """\
-            @@ -36,3 +36,2 @@
-               "requirements": [
-            -    "ansicolors"
-               ],
-            """
-        ),
-        match=re.escape("Expected '.requirements' in <string> to have at least one requirement."),
-    )
-
-
 def test_load_invalid_no_artifacts(patch_tool):
     # type: (PatchTool) -> None
 
@@ -484,7 +479,7 @@ def test_load_invalid_no_artifacts(patch_tool):
             -            {
             -              "algorithm": "md5",
             -              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
-            -              "url": "http://localhost:9999/ansicolors-1.1.8-py2.py3-none-any.whl"
+            -              "url": "file:///find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
             -            },
             -            {
             -              "algorithm": "md5",
@@ -500,81 +495,109 @@ def test_load_invalid_no_artifacts(patch_tool):
     )
 
 
-def test_load_invalid_no_locked_requirements(patch_tool):
+def test_path_mappings_required_on_parse(patch_tool):
     # type: (PatchTool) -> None
 
-    assert_parse_error(
-        patch_tool,
-        dedent(
-            """\
-            @@ -8,20 +8,2 @@
-                   "locked_requirements": [
-            -        {
-            -          "artifacts": [
-            -            {
-            -              "algorithm": "md5",
-            -              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
-            -              "url": "http://localhost:9999/ansicolors-1.1.8-py2.py3-none-any.whl"
-            -            },
-            -            {
-            -              "algorithm": "md5",
-            -              "hash": "9ca7e2396ffa2e20af023c6b83ab7b14",
-            -              "url": "http://localhost:9999/ansicolors-1.1.8.zip"
-            -            }
-            -          ],
-            -          "project_name": "ansicolors",
-            -          "requires_dists": [],
-            -          "requires_python": null,
-            -          "version": "1.1.8"
-            -        }
-                   ],
-            """
-        ),
-        match=re.escape(
-            "Expected '.locked_resolves[0][\"locked_requirements\"]' in <string> to have at least "
-            "one locked requirement."
-        ),
+    patch = dedent(
+        """\
+        @@ -35,2 +35,6 @@
+           ],
+        +  "path_mappings": {
+        +    "FOO": "The fooey find links repo path on the local machine.",
+        +    "BAR": null
+        +  },
+           "pex_version": "2.1.50",
+        """
+    )
+
+    def assert_mapping_error(
+        unspecified_paths,  # type: Container[str]
+        *path_mappings  # type: PathMapping
+    ):
+        with pytest.raises(PathMappingError) as exec_info:
+            parse_patched(patch_tool, patch, path_mappings=PathMappings(path_mappings))
+        assert (
+            PathMappingError(
+                required_path_mappings={
+                    "BAR": None,
+                    "FOO": "The fooey find links repo path on the local machine.",
+                },
+                unspecified_paths=unspecified_paths,
+            )
+            == exec_info.value
+        )
+
+    assert_mapping_error({"BAR", "FOO"})
+    assert_mapping_error({"FOO"}, PathMapping(name="BAR", path="/tmp/bar"))
+    assert_mapping_error({"BAR"}, PathMapping(name="FOO", path="/tmp/foo"))
+
+    assert (
+        parse_patched(
+            patch_tool,
+            patch,
+            path_mappings=PathMappings(
+                (PathMapping(name="FOO", path="/tmp/foo"), PathMapping(name="BAR", path="/tmp/bar"))
+            ),
+        )
+        is not None
     )
 
 
-def test_load_invalid_no_locked_resolves(patch_tool):
-    # type: (PatchTool) -> None
+def test_path_mappings_round_trip():
+    # type: () -> None
 
-    assert_parse_error(
-        patch_tool,
-        dedent(
-            """\
-            @@ -6,29 +6,2 @@
-               "locked_resolves": [
-            -    {
-            -      "locked_requirements": [
-            -        {
-            -          "artifacts": [
-            -            {
-            -              "algorithm": "md5",
-            -              "hash": "f357aa02db2466bc24ff1815cff1aeb3",
-            -              "url": "http://localhost:9999/ansicolors-1.1.8-py2.py3-none-any.whl"
-            -            },
-            -            {
-            -              "algorithm": "md5",
-            -              "hash": "9ca7e2396ffa2e20af023c6b83ab7b14",
-            -              "url": "http://localhost:9999/ansicolors-1.1.8.zip"
-            -            }
-            -          ],
-            -          "project_name": "ansicolors",
-            -          "requires_dists": [],
-            -          "requires_python": null,
-            -          "version": "1.1.8"
-            -        }
-            -      ],
-            -      "platform_tag": [
-            -        "cp39",
-            -        "cp39",
-            -        "manylinux_2_33_x86_64"
-            -      ]
-            -    }
-               ],
-            """
+    lock_file = json_codec.loads(VALID_LOCK)
+
+    data = json_codec.as_json_data(
+        lock_file,
+        PathMappings(
+            (
+                PathMapping(
+                    path="/find-links",
+                    name="FL",
+                    description="Our NFS find-links repo local mount path.",
+                ),
+            )
         ),
-        match=re.escape("Expected '.locked_resolves' in <string> to have at least one resolve."),
+    )
+    locked_resolves = data["locked_resolves"]
+    assert 1 == len(locked_resolves)
+    locked_requirements = locked_resolves[0]["locked_requirements"]
+    assert 1 == len(locked_requirements)
+    artifacts = locked_requirements[0]["artifacts"]
+    assert 2 == len(artifacts)
+    assert "file://${FL}/ansicolors-1.1.8-py2.py3-none-any.whl" == artifacts[0]["url"]
+    assert "http://localhost:9999/ansicolors-1.1.8.zip" == artifacts[1]["url"]
+
+    canonicalized_lock = json.dumps(data)
+    with pytest.raises(PathMappingError) as exc_info:
+        json_codec.loads(canonicalized_lock)
+    assert (
+        PathMappingError(
+            required_path_mappings={"FL": "Our NFS find-links repo local mount path."},
+            unspecified_paths={"FL"},
+        )
+        == exc_info.value
+    )
+
+    assert lock_file == json_codec.loads(
+        canonicalized_lock,
+        path_mappings=PathMappings((PathMapping(name="FL", path="/find-links"),)),
+    ), (
+        "Expected a round-trip through canonicalize / reify with the same mappings to return the "
+        "same lock contents."
+    )
+
+    lock_file2 = json_codec.loads(
+        canonicalized_lock,
+        path_mappings=PathMappings((PathMapping(name="FL", path="/at/another/path/find-links"),)),
+    )
+    assert (
+        lock_file2 != lock_file
+    ), "Expected path roots to be reified to different values than we started with."
+    assert (
+        ArtifactURL.parse(
+            "file:///at/another/path/find-links/ansicolors-1.1.8-py2.py3-none-any.whl"
+        )
+        == lock_file2.locked_resolves[0].locked_requirements[0].artifact.url
     )
