@@ -1,10 +1,11 @@
-# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2014 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 import json
 import os
-import subprocess
+import site
 import sys
+import sysconfig
 import tempfile
 import textwrap
 from contextlib import contextmanager
@@ -14,29 +15,33 @@ from types import ModuleType
 import pytest
 
 from pex import resolver
-from pex.common import safe_mkdir, safe_open, temporary_dir
-from pex.compatibility import PY2, WINDOWS, to_bytes
-from pex.interpreter import PythonInterpreter
-from pex.pex import PEX
+from pex.common import environment_as, safe_mkdir, safe_open, temporary_dir
+from pex.compatibility import PY2, safe_commonpath, to_bytes
+from pex.dist_metadata import Distribution
+from pex.fs import safe_symlink
+from pex.interpreter import PythonIdentity, PythonInterpreter
+from pex.os import WINDOWS
+from pex.pex import PEX, IsolatedSysPath
 from pex.pex_builder import PEXBuilder
 from pex.pex_info import PexInfo
-from pex.testing import (
-    IS_PYPY3,
-    PY27,
+from pex.resolve.configured_resolver import ConfiguredResolver
+from pex.typing import TYPE_CHECKING
+from pex.util import named_temporary_file
+from pex.version import __version__
+from testing import (
+    PY39,
     PY310,
+    PY_VER,
     WheelBuilder,
-    built_wheel,
     ensure_python_interpreter,
-    environment_as,
+    install_wheel,
     make_bdist,
     run_simple_pex,
     run_simple_pex_test,
+    subprocess,
     temporary_content,
     write_simple_pex,
 )
-from pex.third_party.pkg_resources import Distribution
-from pex.typing import TYPE_CHECKING
-from pex.util import named_temporary_file
 
 try:
     from unittest import mock
@@ -44,7 +49,7 @@ except ImportError:
     import mock  # type: ignore[no-redef,import]
 
 if TYPE_CHECKING:
-    from typing import Dict, Iterable, Iterator, Union
+    from typing import Any, Dict, Iterable, Iterator, Mapping, Optional, Union
 
     import attr  # vendor:skip
 else:
@@ -67,7 +72,6 @@ def test_excepthook_honored():
 
         def excepthook(ex_type, ex, tb):
             print('Custom hook called with: {0}'.format(ex))
-            sys.exit(42)
 
         sys.excepthook = excepthook
 
@@ -77,7 +81,7 @@ def test_excepthook_honored():
 
     so, rc = run_simple_pex_test(body)
     assert so == b"Custom hook called with: This is an exception\n", "Standard out was: %r" % so
-    assert rc == 42
+    assert rc == 1
 
 
 def _test_sys_exit(arg, expected_output, expected_rc):
@@ -113,7 +117,6 @@ def test_pex_sys_exit_prints_objects():
     _test_sys_exit('Exception("derp")', b"derp\n", 1)
 
 
-@pytest.mark.xfail(IS_PYPY3, reason="https://github.com/pantsbuild/pex/issues/1210")
 def test_pex_atexit_swallowing():
     # type: () -> None
     body = textwrap.dedent(
@@ -128,25 +131,30 @@ def test_pex_atexit_swallowing():
     )
 
     so, rc = run_simple_pex_test(body)
-    assert so == b""
-    assert rc == 0
-
-    env_copy = os.environ.copy()
-    env_copy.update(PEX_TEARDOWN_VERBOSE="1")
-    so, rc = run_simple_pex_test(body, env=env_copy)
     assert b"This is an exception" in so
     assert rc == 0
 
 
 def test_minimum_sys_modules():
     # type: () -> None
+
+    def minimum_sys_modules(
+        sys_path=(),  # type: Iterable[str]
+        site_libs=(),  # type: Iterable[str]
+        modules=None,  # type: Optional[Mapping[str, ModuleType]]
+    ):
+        # type: (...) -> Mapping[str, ModuleType]
+        return PEX.minimum_sys_modules(
+            IsolatedSysPath(sys_path=sys_path, site_packages=site_libs), modules=modules
+        )
+
     # tainted modules evict
     tainted_module = ModuleType("tainted_module")
     tainted_module.__file__ = "bad_path"
     modules = {"tainted_module": tainted_module}
-    new_modules = PEX.minimum_sys_modules(site_libs=[], modules=modules)
+    new_modules = minimum_sys_modules(sys_path=["bad_path"], modules=modules)
     assert new_modules == modules
-    new_modules = PEX.minimum_sys_modules(site_libs=["bad_path"], modules=modules)
+    new_modules = minimum_sys_modules(site_libs=["bad_path"], modules=modules)
     assert new_modules == {}
 
     # builtins stay
@@ -154,18 +162,20 @@ def test_minimum_sys_modules():
     stdlib_module = ModuleType("my_stdlib")
     stdlib_module.__file__ = "good_path"
     modules = {"my_builtin": builtin_module, "my_stdlib": stdlib_module}
-    new_modules = PEX.minimum_sys_modules(site_libs=[], modules=modules)
+    new_modules = minimum_sys_modules(sys_path=["good_path"], modules=modules)
     assert new_modules == modules
-    new_modules = PEX.minimum_sys_modules(site_libs=["bad_path"], modules=modules)
+    new_modules = minimum_sys_modules(
+        sys_path=["good_path"], site_libs=["bad_path"], modules=modules
+    )
     assert new_modules == modules
 
     # tainted packages evict
     tainted_module = ModuleType("tainted_module")
     tainted_module.__path__ = ["bad_path"]  # type: ignore[attr-defined]
     modules = {"tainted_module": tainted_module}
-    new_modules = PEX.minimum_sys_modules(site_libs=[], modules=modules)
+    new_modules = minimum_sys_modules(sys_path=["bad_path"], modules=modules)
     assert new_modules == modules
-    new_modules = PEX.minimum_sys_modules(site_libs=["bad_path"], modules=modules)
+    new_modules = minimum_sys_modules(site_libs=["bad_path"], modules=modules)
     assert new_modules == {}
     assert tainted_module.__path__ == []  # type: ignore[attr-defined]
 
@@ -173,14 +183,16 @@ def test_minimum_sys_modules():
     tainted_module = ModuleType("tainted_module")
     tainted_module.__path__ = ["bad_path", "good_path"]  # type: ignore[attr-defined]
     modules = {"tainted_module": tainted_module}
-    new_modules = PEX.minimum_sys_modules(site_libs=[], modules=modules)
+    new_modules = minimum_sys_modules(sys_path=["good_path"], site_libs=[], modules=modules)
     assert new_modules == modules
-    new_modules = PEX.minimum_sys_modules(site_libs=["bad_path"], modules=modules)
+    new_modules = minimum_sys_modules(
+        sys_path=["good_path"], site_libs=["bad_path"], modules=modules
+    )
     assert new_modules == modules
     assert tainted_module.__path__ == ["good_path"]  # type: ignore[attr-defined]
 
     # If __path__ is not a list the module is removed; typically this implies
-    # it's a namespace package (https://www.python.org/dev/peps/pep-0420/) where
+    # it's a namespace package (https://peps.python.org/pep-0420/) where
     # __path__ is a _NamespacePath.
     try:
         from importlib._bootstrap_external import _NamespacePath  # type: ignore
@@ -195,44 +207,63 @@ def test_minimum_sys_modules():
     tainted_module = FakeModule()  # type: ignore[assignment]
     tainted_module.__path__ = bad_path  # type: ignore[attr-defined] # Not a list as expected
     modules = {"tainted_module": tainted_module}
-    new_modules = PEX.minimum_sys_modules(site_libs=["bad_path"], modules=modules)
+    new_modules = minimum_sys_modules(site_libs=["bad_path"], modules=modules)
     assert new_modules == {}
 
     # If __file__ is explicitly None we should gracefully proceed to __path__ checks.
     tainted_module = ModuleType("tainted_module")
     tainted_module.__file__ = None  # type: ignore[assignment]
     modules = {"tainted_module": tainted_module}
-    new_modules = PEX.minimum_sys_modules(site_libs=[], modules=modules)
+    new_modules = minimum_sys_modules(site_libs=[], modules=modules)
     assert new_modules == modules
 
 
-def test_site_libs():
-    # type: () -> None
-    with mock.patch.object(
-        PEX, "_get_site_packages"
-    ) as mock_site_packages, temporary_dir() as tempdir:
-        site_packages = os.path.join(tempdir, "site-packages")
+def test_site_libs(tmpdir):
+    # type: (Any) -> None
+    with mock.patch.object(site, "getsitepackages", create=True) as mock_site_packages:
+        site_packages = os.path.join(str(tmpdir), "site-packages")
         os.mkdir(site_packages)
-        mock_site_packages.return_value = set([site_packages])
-        site_libs = PEX.site_libs()
+        mock_site_packages.return_value = {site_packages}
+        site_libs = frozenset(entry.path for entry in PythonIdentity.get().site_packages)
         assert site_packages in site_libs
 
 
 @pytest.mark.skipif(WINDOWS, reason="No symlinks on windows")
-def test_site_libs_symlink():
-    # type: () -> None
-    with mock.patch.object(
-        PEX, "_get_site_packages"
-    ) as mock_site_packages, temporary_dir() as tempdir:
-        site_packages = os.path.join(tempdir, "site-packages")
-        os.mkdir(site_packages)
-        site_packages_link = os.path.join(tempdir, "site-packages-link")
-        os.symlink(site_packages, site_packages_link)
-        mock_site_packages.return_value = set([site_packages_link])
+def test_site_libs_symlink(tmpdir):
+    # type: (Any) -> None
 
-        site_libs = PEX.site_libs()
-        assert os.path.realpath(site_packages) in site_libs
-        assert site_packages_link in site_libs
+    # N.B.: PythonIdentity.get() used below in the core test ends up consulting
+    # `packaging.tags.sys_tags()` which in turn consults `sysconfig`; so we need to make sure that
+    # bit of stdlib is on the sys.path. We get this by grabbing its sys.path entry, which contains
+    # the whole stdlib in addition to it.
+    assert sysconfig.__file__
+
+    sys_path_entries = tuple(
+        entry
+        for entry in PythonIdentity.get().sys_path
+        if entry == safe_commonpath((entry, sysconfig.__file__))
+    )
+    assert len(sys_path_entries) == 1
+    sys_path_entry = sys_path_entries[0]
+
+    sys_path_entry_link = os.path.join(str(tmpdir), "lib-link")
+    safe_symlink(sys_path_entry, sys_path_entry_link)
+
+    with mock.patch.object(site, "getsitepackages", create=True) as mock_site_packages, mock.patch(
+        "sys.path", new=[sys_path_entry_link]
+    ):
+        site_packages = os.path.join(str(tmpdir), "site-packages")
+        os.mkdir(site_packages)
+        site_packages_link = os.path.join(str(tmpdir), "site-packages-link")
+        safe_symlink(site_packages, site_packages_link)
+        mock_site_packages.return_value = [site_packages_link]
+
+        isolated_sys_path = IsolatedSysPath.for_pex(
+            interpreter=PythonIdentity.get(), pex=os.devnull
+        )
+        assert os.path.join(sys_path_entry, "module.py") in isolated_sys_path
+        assert os.path.realpath(site_packages) not in isolated_sys_path
+        assert site_packages_link not in isolated_sys_path
 
 
 def test_site_libs_excludes_prefix():
@@ -243,12 +274,12 @@ def test_site_libs_excludes_prefix():
     """
 
     with mock.patch.object(
-        PEX, "_get_site_packages"
+        site, "getsitepackages", create=True
     ) as mock_site_packages, temporary_dir() as tempdir:
-        site_packages = os.path.join(tempdir, "site-packages")
+        site_packages = os.path.realpath(os.path.join(tempdir, "site-packages"))
         os.mkdir(site_packages)
-        mock_site_packages.return_value = set([site_packages, sys.prefix])
-        site_libs = PEX.site_libs()
+        mock_site_packages.return_value = [site_packages, sys.prefix]
+        site_libs = tuple(entry.path for entry in PythonIdentity.get().site_packages)
         assert site_packages in site_libs
         assert sys.prefix not in site_libs
 
@@ -257,19 +288,19 @@ def test_site_libs_excludes_prefix():
 @pytest.mark.parametrize("project_name", ("my_project", "my-project"))
 def test_pex_script(project_name, zip_safe):
     # type: (str, bool) -> None
-    with built_wheel(name=project_name, zip_safe=zip_safe) as bdist_path:
+    with make_bdist(name=project_name, zip_safe=zip_safe) as bdist:
         env_copy = os.environ.copy()
         env_copy["PEX_SCRIPT"] = "hello_world"
         so, rc = run_simple_pex_test("", env=env_copy)
         assert rc == 1, so.decode("utf-8")
         assert b"Could not find script 'hello_world'" in so
 
-        so, rc = run_simple_pex_test("", env=env_copy, dists=[bdist_path])
+        so, rc = run_simple_pex_test("", env=env_copy, dists=[bdist])
         assert rc == 0, so.decode("utf-8")
         assert b"hello world" in so
 
         env_copy["PEX_SCRIPT"] = "shell_script"
-        so, rc = run_simple_pex_test("", env=env_copy, dists=[bdist_path])
+        so, rc = run_simple_pex_test("", env=env_copy, dists=[bdist])
         assert rc == 0, so.decode("utf-8")
         assert b"hello world from shell script" in so
 
@@ -294,20 +325,24 @@ def test_pex_run_extra_sys_path():
     with named_temporary_file() as fake_stdout:
         with temporary_dir() as temp_dir:
             pex = write_simple_pex(
-                temp_dir, 'import sys; sys.stdout.write(":".join(sys.path)); sys.exit(0)'
+                temp_dir, "import os, sys; sys.stdout.write(os.pathsep.join(sys.path)); sys.exit(0)"
             )
             rc = PEX(pex.path()).run(
                 stdin=None,
                 stdout=fake_stdout,
                 stderr=None,
-                env={"PEX_EXTRA_SYS_PATH": "extra/syspath/entry1:extra/syspath/entry2"},
+                env={
+                    "PEX_EXTRA_SYS_PATH": os.pathsep.join(
+                        ("extra/syspath/entry1", "extra/syspath/entry2")
+                    )
+                },
             )
             assert rc == 0
 
             fake_stdout.seek(0)
-            syspath = fake_stdout.read().split(b":")
-            assert b"extra/syspath/entry1" in syspath
-            assert b"extra/syspath/entry2" in syspath
+            syspath = fake_stdout.read().decode("utf-8").split(os.pathsep)
+            assert "extra/syspath/entry1" in syspath
+            assert "extra/syspath/entry2" in syspath
 
 
 @attr.s(frozen=True)
@@ -324,26 +359,36 @@ class PythonpathIsolationTest(object):
     def assert_isolation(self, inherit_path, expected_output):
         # type: (Union[str, bool], str) -> None
         env = dict(PYTHONPATH=self.pythonpath)
-        with named_temporary_file() as fake_stdout:
-            with temporary_dir() as temp_dir:
-                pex_builder = write_simple_pex(
-                    temp_dir,
-                    pex_info=self.pex_info(inherit_path),
-                    dists=self.dists,
-                    exe_contents=self.exe,
-                )
+        with temporary_dir() as temp_dir:
+            pex_builder = write_simple_pex(
+                temp_dir,
+                pex_info=self.pex_info(inherit_path),
+                dists=self.dists,
+                exe_contents=self.exe,
+            )
 
-                # Test the PEX.run API.
-                rc = PEX(pex_builder.path()).run(stdout=fake_stdout, env=env)
-                assert rc == 0
+            interpreter = PythonInterpreter.get()
+            if PY2:
+                # Under Python 2 venvs (created by Pex via Virtualenv 16.7.12), the venv interpreter
+                # can fail to resolve stdlib weakref internals. We side-step by just resolving out
+                # of the venv since this test is aimed squarely at PYTHONPATH isolation.
+                interpreter = interpreter.resolve_base_interpreter()
 
-                fake_stdout.seek(0)
-                assert expected_output == fake_stdout.read().decode("utf-8")
+            # Test the PEX.run API.
+            process = PEX(pex_builder.path(), interpreter=interpreter).run(
+                stdout=subprocess.PIPE,
+                stderr=subprocess.PIPE,
+                env=env,
+                blocking=False,
+            )
+            stdout, stderr = process.communicate()
+            assert process.returncode == 0, stderr.decode("utf-8")
+            assert expected_output == stdout.decode("utf-8")
 
-                # Test direct PEX execution.
-                assert expected_output == subprocess.check_output(
-                    [sys.executable, pex_builder.path()], env=env
-                ).decode("utf-8")
+            # Test direct PEX execution.
+            assert expected_output == subprocess.check_output(
+                [interpreter.binary, pex_builder.path()], env=env
+            ).decode("utf-8")
 
 
 @pytest.fixture(scope="module")
@@ -374,8 +419,7 @@ def pythonpath_isolation_test():
         }
 
         with temporary_content(dist_content) as project_dir:
-            installer = WheelBuilder(project_dir)
-            foo_bdist = installer.bdist()
+            foo_bdist = install_wheel(WheelBuilder(project_dir).bdist())
 
             exe_contents = textwrap.dedent(
                 """
@@ -472,9 +516,8 @@ def test_pex_executable():
         }  # type: Dict[str, Union[str, int]]
         pex_builder = PEXBuilder(path=pex_dir)
         with temporary_content(project_content, perms=0o755) as project_dir:
-            installer = WheelBuilder(project_dir)
-            bdist = installer.bdist()
-            pex_builder.add_dist_location(bdist)
+            bdist = install_wheel(WheelBuilder(project_dir).bdist())
+            pex_builder.add_dist_location(bdist.location)
             pex_builder.set_executable(os.path.join(pex_dir, "exe.py"))
             pex_builder.freeze()
 
@@ -567,11 +610,11 @@ def test_pex_verify_entry_point_module_should_fail():
 def test_activate_interpreter_different_from_current():
     # type: () -> None
     with temporary_dir() as pex_root:
-        interp_version = PY310 if PY2 else PY27
+        interp_version = PY310 if PY_VER == (3, 9) else PY39
         custom_interpreter = PythonInterpreter.from_binary(
             ensure_python_interpreter(interp_version)
         )
-        pex_info = PexInfo.default(custom_interpreter)
+        pex_info = PexInfo.default()
         pex_info.pex_root = pex_root
         with temporary_dir() as pex_chroot:
             pex_builder = PEXBuilder(
@@ -607,6 +650,27 @@ def test_execute_interpreter_dashc_program():
         assert b"" == stderr
 
 
+def test_execute_interpreter_dashc_program_with_python_options():
+    with temporary_dir() as pex_chroot:
+        pex_builder = PEXBuilder(path=pex_chroot)
+        pex_builder.freeze()
+        # To test with interpreter options we add an "assert False"
+        # that we would expect to fail.
+        # adding the -O option will ignore that assertion
+        # see: https://docs.python.org/3/using/cmdline.html#cmdoption-O
+        process = PEX(pex_chroot).run(
+            args=["-O", "-c", 'assert False; import sys; print(" ".join(sys.argv))', "one"],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            blocking=False,
+        )
+        stdout, stderr = process.communicate()
+
+        assert b"" == stderr
+        assert b"-c one\n" == stdout
+        assert 0 == process.returncode
+
+
 def test_execute_interpreter_dashm_module():
     # type: () -> None
     with temporary_dir() as pex_chroot:
@@ -642,6 +706,46 @@ def test_execute_interpreter_dashm_module():
         assert b"" == stderr
 
 
+def test_execute_interpreter_dashm_module_with_python_options():
+    # type: () -> None
+    with temporary_dir() as pex_chroot:
+        pex_builder = PEXBuilder(path=pex_chroot)
+        pex_builder.add_source(None, "foo/__init__.py")
+        # To test with interpreter options we add an "assert False"
+        # that we would expect to fail.
+        # adding the -O option will ignore that assertion
+        # see: https://docs.python.org/3/using/cmdline.html#cmdoption-O
+        with tempfile.NamedTemporaryFile(mode="w") as fp:
+            fp.write(
+                dedent(
+                    """\
+                    import os
+                    import sys
+
+                    assert False
+                    print("{} {}".format(os.path.realpath(sys.argv[0]), " ".join(sys.argv[1:])))
+                    """
+                )
+            )
+            fp.flush()
+            pex_builder.add_source(fp.name, "foo/bar.py")
+        pex_builder.freeze()
+        pex = PEX(pex_chroot)
+        process = pex.run(
+            args=["-O", "-m", "foo.bar", "one", "two"],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            blocking=False,
+        )
+        stdout, stderr = process.communicate()
+
+        assert b"" == stderr
+        assert "{} one two\n".format(
+            os.path.realpath(os.path.join(pex_chroot, "foo/bar.py"))
+        ) == stdout.decode("utf-8")
+        assert 0 == process.returncode
+
+
 def test_execute_interpreter_stdin_program():
     # type: () -> None
     with temporary_dir() as pex_chroot:
@@ -661,6 +765,31 @@ def test_execute_interpreter_stdin_program():
         assert b"" == stderr
 
 
+def test_execute_interpreter_stdin_program_with_python_options():
+    # type: () -> None
+    with temporary_dir() as pex_chroot:
+        pex_builder = PEXBuilder(path=pex_chroot)
+        pex_builder.freeze()
+        # To test with interpreter options we add an "assert False"
+        # that we would expect to fail.
+        # adding the -O option will ignore that assertion
+        # see: https://docs.python.org/3/using/cmdline.html#cmdoption-O
+        process = PEX(pex_chroot).run(
+            args=["-O", "-", "one", "two"],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            stdin=subprocess.PIPE,
+            blocking=False,
+        )
+        stdout, stderr = process.communicate(
+            input=b'import sys; assert False; print(" ".join(sys.argv))'
+        )
+
+        assert b"" == stderr
+        assert b"- one two\n" == stdout
+        assert 0 == process.returncode
+
+
 def test_execute_interpreter_file_program():
     # type: () -> None
     with temporary_dir() as pex_chroot:
@@ -682,6 +811,76 @@ def test_execute_interpreter_file_program():
             assert b"" == stderr
 
 
+EXPECTED_REPL_BANNER_NO_DEPS = (
+    dedent(
+        """\
+        Pex {pex_version} hermetic environment with no dependencies.
+        Python {python_version} on {platform}
+        Type "help", "pex", "copyright", "credits" or "license" for more information.
+        """
+    )
+    .format(python_version=sys.version, platform=sys.platform, pex_version=__version__)
+    .encode("utf-8")
+)
+
+
+def test_execute_repl():
+    with temporary_dir() as pex_chroot:
+        pex_builder = PEXBuilder(path=pex_chroot)
+        pex_builder.freeze()
+        process = PEX(pex_chroot).run(
+            args=[],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            stdin=subprocess.PIPE,
+            blocking=False,
+        )
+        commands = dedent(
+            """
+            assert False
+            20 + 103
+            quit()
+            """
+        )
+        stdout, stderr = process.communicate(input=commands.encode("utf-8"))
+
+        assert stderr.startswith(EXPECTED_REPL_BANNER_NO_DEPS), stderr
+        assert b"AssertionError" in stderr
+        assert b">>> " in stdout
+        assert b"123" in stdout
+        assert 0 == process.returncode
+
+
+def test_execute_repl_with_python_options():
+    with temporary_dir() as pex_chroot:
+        pex_builder = PEXBuilder(path=pex_chroot)
+        pex_builder.freeze()
+        process = PEX(pex_chroot).run(
+            args=["-O"],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            stdin=subprocess.PIPE,
+            blocking=False,
+        )
+        # adding the -O option will ignore that assertion
+        # see: https://docs.python.org/3/using/cmdline.html#cmdoption-O
+        # We should not see the AssertionError in the output anymore.
+        commands = dedent(
+            """
+            assert False
+            20 + 103
+            quit()
+            """
+        )
+        stdout, stderr = process.communicate(input=commands.encode("utf-8"))
+
+        assert stderr.startswith(EXPECTED_REPL_BANNER_NO_DEPS), stderr
+        assert b"AssertionError" not in stderr
+        assert b">>> " in stdout
+        assert b"123" in stdout
+        assert 0 == process.returncode
+
+
 def test_pex_run_strip_env():
     # type: () -> None
     with temporary_dir() as pex_root:
@@ -713,27 +912,53 @@ def test_pex_run_strip_env():
             }, "Expected the parent environment to be left un-stripped."
 
 
-def test_pex_run_custom_setuptools_useable():
-    # type: () -> None
-    result = resolver.resolve(requirements=["setuptools==43.0.0"])
-    dists = [installed_dist.distribution for installed_dist in result.installed_distributions]
+@pytest.fixture
+def setuptools_version():
+    # type: () -> str
+    return "67.8.0" if sys.version_info[:2] >= (3, 12) else "43.0.0"
+
+
+@pytest.fixture
+def setuptools_requirement(setuptools_version):
+    # type: (str) -> str
+    return "setuptools=={version}".format(version=setuptools_version)
+
+
+def test_pex_run_custom_setuptools_useable(
+    setuptools_requirement,  # type: str
+    setuptools_version,  # type: str
+):
+    # type: (...) -> None
+    result = resolver.resolve(
+        requirements=[setuptools_requirement],
+        resolver=ConfiguredResolver.default(),
+    )
+    dists = [resolved_dist.distribution for resolved_dist in result.distributions]
     with temporary_dir() as temp_dir:
         pex = write_simple_pex(
             temp_dir,
-            "import setuptools, sys; sys.exit(0 if '43.0.0' == setuptools.__version__ else 1)",
+            "import setuptools, sys; sys.exit(0 if '{version}' == setuptools.__version__ else 1)".format(
+                version=setuptools_version
+            ),
             dists=dists,
         )
         rc = PEX(pex.path()).run()
         assert rc == 0
 
 
-def test_pex_run_conflicting_custom_setuptools_useable():
-    # type: () -> None
+def test_pex_run_conflicting_custom_setuptools_useable(
+    setuptools_requirement,  # type: str
+    setuptools_version,  # type: str
+):
+    # type: (...) -> None
     # Here we use our vendored, newer setuptools to build the pex which has an older setuptools
     # requirement.
 
-    result = resolver.resolve(requirements=["setuptools==43.0.0"])
-    dists = [installed_dist.distribution for installed_dist in result.installed_distributions]
+    result = resolver.resolve(
+        requirements=[setuptools_requirement],
+        resolver=ConfiguredResolver.default(),
+    )
+    dists = [resolved_dist.distribution for resolved_dist in result.distributions]
     with temporary_dir() as temp_dir:
         pex = write_simple_pex(
             temp_dir,
@@ -742,8 +967,10 @@ def test_pex_run_conflicting_custom_setuptools_useable():
                 import sys
                 import setuptools
 
-                sys.exit(0 if '43.0.0' == setuptools.__version__ else 1)
-                """
+                sys.exit(0 if '{version}' == setuptools.__version__ else 1)
+                """.format(
+                    version=setuptools_version
+                )
             ),
             dists=dists,
         )
@@ -755,9 +982,10 @@ def test_pex_run_custom_pex_useable():
     # type: () -> None
     old_pex_version = "0.7.0"
     result = resolver.resolve(
-        requirements=["pex=={}".format(old_pex_version), "setuptools==40.6.3"]
+        requirements=["pex=={}".format(old_pex_version), "setuptools==40.6.3"],
+        resolver=ConfiguredResolver.default(),
     )
-    dists = [installed_dist.distribution for installed_dist in result.installed_distributions]
+    dists = [resolved_dist.distribution for resolved_dist in result.distributions]
     with temporary_dir() as temp_dir:
         from pex.version import __version__
 
@@ -773,8 +1001,13 @@ def test_pex_run_custom_pex_useable():
                   from pex.version import __version__
                   sys.exit(1)
                 except ImportError:
-                  import pkg_resources
-                  dist = pkg_resources.working_set.find(pkg_resources.Requirement.parse('pex'))
+                  # N.B.: pkg_resources is not supported by Python >= 3.12.
+                  if sys.version_info[:2] >= (3, 12):
+                      from importlib.metadata import distribution
+                      dist = distribution('pex')
+                  else:
+                      import pkg_resources
+                      dist = pkg_resources.working_set.find(pkg_resources.Requirement.parse('pex'))
                   print(dist.version)
                 """
             ),
@@ -791,7 +1024,7 @@ def test_pex_run_custom_pex_useable():
 def test_interpreter_teardown_dev_null_unclosed_resource_warning_suppressed():
     # type: () -> None
 
-    # See https://github.com/pantsbuild/pex/issues/1101 and
+    # See https://github.com/pex-tool/pex/issues/1101 and
     # https://github.com/pantsbuild/pants/issues/11058 for the motivating issue.
     with temporary_dir() as pex_chroot:
         pex_builder = PEXBuilder(path=pex_chroot)
