@@ -1,40 +1,48 @@
-# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2015 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 import os
-import subprocess
+import shutil
 import sys
+import zipfile
 from collections import defaultdict
+from contextlib import contextmanager
 from textwrap import dedent
 
-import pkginfo
 import pytest
 
-from pex import targets
+from pex import dist_metadata, targets
+from pex.build_system.pep_517 import build_sdist
 from pex.common import safe_copy, safe_mkdtemp, temporary_dir
-from pex.interpreter import PythonInterpreter, spawn_python_job
-from pex.platforms import Platform
-from pex.resolve.resolver_configuration import ResolverVersion
-from pex.resolve.resolvers import InstalledDistribution, Unsatisfiable
-from pex.resolver import IntegrityError, LocalDistribution, download, install, resolve
-from pex.targets import Targets
-from pex.testing import (
+from pex.dist_metadata import Distribution, Requirement
+from pex.interpreter import PythonInterpreter
+from pex.pip.version import PipVersion
+from pex.resolve import abbreviated_platforms
+from pex.resolve.configured_resolver import ConfiguredResolver
+from pex.resolve.resolver_configuration import BuildConfiguration, ResolverVersion
+from pex.resolve.resolvers import ResolvedDistribution, ResolveResult, Unsatisfiable
+from pex.resolver import download
+from pex.resolver import resolve as resolve_under_test
+from pex.targets import Target, Targets
+from pex.typing import TYPE_CHECKING
+from pex.variables import ENV
+from testing import (
     IS_LINUX,
     IS_PYPY,
     PY27,
-    PY37,
+    PY39,
     PY310,
+    PY311,
     PY_VER,
     built_wheel,
     ensure_python_interpreter,
     make_project,
     make_source_dir,
+    subprocess,
 )
-from pex.third_party.pkg_resources import Requirement
-from pex.typing import TYPE_CHECKING
 
 if TYPE_CHECKING:
-    from typing import Any, Iterable, List, Union
+    from typing import Any, DefaultDict, Iterable, Iterator, List, Union
 
 
 def create_sdist(**kwargs):
@@ -42,14 +50,12 @@ def create_sdist(**kwargs):
     dist_dir = safe_mkdtemp()
 
     with make_project(**kwargs) as project_dir:
-        cmd = ["setup.py", "sdist", "--dist-dir={}".format(dist_dir)]
-        spawn_python_job(
-            args=cmd,
-            cwd=project_dir,
-            expose=["setuptools"],
-            stdout=subprocess.PIPE,
-            stderr=subprocess.PIPE,
-        ).communicate()
+        build_sdist(
+            project_directory=project_dir,
+            dist_dir=dist_dir,
+            target=targets.current(),
+            resolver=ConfiguredResolver.default(),
+        )
 
     dists = os.listdir(dist_dir)
     assert len(dists) == 1
@@ -62,11 +68,34 @@ def build_wheel(**kwargs):
         return whl
 
 
+def resolve(**kwargs):
+    # type: (**Any) -> ResolveResult
+    kwargs.setdefault("resolver", ConfiguredResolver.default())
+    return resolve_under_test(**kwargs)
+
+
 def local_resolve(*args, **kwargs):
-    # type: (*Any, **Any) -> List[InstalledDistribution]
+    # type: (*Any, **Any) -> List[ResolvedDistribution]
     # Skip remote lookups.
     kwargs["indexes"] = []
-    return list(resolve(*args, **kwargs).installed_distributions)
+    return list(resolve(*args, **kwargs).distributions)
+
+
+@contextmanager
+def cache(directory):
+    # type: (str) -> Iterator[None]
+    with ENV.patch(PEX_ROOT=directory):
+        yield
+
+
+@contextmanager
+def disabled_cache():
+    # type: () -> Iterator[None]
+
+    # N.B.: The resolve cache is never actually disabled, `--disable-cache` just switches the cache
+    # from default PEX_ROOT to a temporary directory. We do the same here.
+    with temporary_dir() as td, cache(td):
+        yield
 
 
 def test_empty_resolve():
@@ -74,8 +103,8 @@ def test_empty_resolve():
     empty_resolve = local_resolve(requirements=[])
     assert empty_resolve == []
 
-    with temporary_dir() as td:
-        empty_resolve = local_resolve(requirements=[], cache=td)
+    with disabled_cache():
+        empty_resolve = local_resolve(requirements=[])
         assert empty_resolve == []
 
 
@@ -85,74 +114,71 @@ def test_simple_local_resolve():
 
     with temporary_dir() as td:
         safe_copy(project_wheel, os.path.join(td, os.path.basename(project_wheel)))
-        installed_dists = local_resolve(requirements=["project"], find_links=[td])
-        assert len(installed_dists) == 1
+        resolved_dists = local_resolve(requirements=["project"], find_links=[td])
+        assert len(resolved_dists) == 1
 
 
 def test_resolve_cache():
     # type: () -> None
     project_wheel = build_wheel(name="project")
 
-    with temporary_dir() as td, temporary_dir() as cache:
+    with temporary_dir() as td, temporary_dir() as cache_dir:
         safe_copy(project_wheel, os.path.join(td, os.path.basename(project_wheel)))
 
         # Without a cache, each resolve should be isolated, but otherwise identical.
-        installed_dists1 = local_resolve(requirements=["project"], find_links=[td])
-        installed_dists2 = local_resolve(requirements=["project"], find_links=[td])
-        assert installed_dists1 != installed_dists2
-        assert len(installed_dists1) == 1
-        assert len(installed_dists2) == 1
-        assert installed_dists1[0].direct_requirements == installed_dists2[0].direct_requirements
-        assert (
-            installed_dists1[0].distribution.location != installed_dists2[0].distribution.location
-        )
+        with disabled_cache():
+            resolved_dists1 = local_resolve(requirements=["project"], find_links=[td])
+        with disabled_cache():
+            resolved_dists2 = local_resolve(requirements=["project"], find_links=[td])
+        assert resolved_dists1 != resolved_dists2
+        assert len(resolved_dists1) == 1
+        assert len(resolved_dists2) == 1
+        assert resolved_dists1[0].direct_requirements == resolved_dists2[0].direct_requirements
+        assert resolved_dists1[0].distribution.location != resolved_dists2[0].distribution.location
 
         # With a cache, each resolve should be identical.
-        installed_dists3 = local_resolve(requirements=["project"], find_links=[td], cache=cache)
-        installed_dists4 = local_resolve(requirements=["project"], find_links=[td], cache=cache)
-        assert installed_dists1 != installed_dists3
-        assert installed_dists2 != installed_dists3
-        assert installed_dists3 == installed_dists4
+        with cache(cache_dir):
+            resolved_dists3 = local_resolve(requirements=["project"], find_links=[td])
+            resolved_dists4 = local_resolve(requirements=["project"], find_links=[td])
+        assert resolved_dists1 != resolved_dists3
+        assert resolved_dists2 != resolved_dists3
+        assert resolved_dists3 == resolved_dists4
 
 
 def test_diamond_local_resolve_cached():
     # type: () -> None
-    # This exercises the issue described here: https://github.com/pantsbuild/pex/issues/120
+    # This exercises the issue described here: https://github.com/pex-tool/pex/issues/120
     project1_wheel = build_wheel(name="project1", install_reqs=["project2<1.0.0"])
     project2_wheel = build_wheel(name="project2")
 
     with temporary_dir() as dd:
         for wheel in (project1_wheel, project2_wheel):
             safe_copy(wheel, os.path.join(dd, os.path.basename(wheel)))
-        with temporary_dir() as cd:
-            installed_dists = local_resolve(
-                requirements=["project1", "project2"], find_links=[dd], cache=cd
-            )
-            assert len(installed_dists) == 2
+        with temporary_dir() as cd, cache(cd):
+            resolved_dists = local_resolve(requirements=["project1", "project2"], find_links=[dd])
+            assert len(resolved_dists) == 2
 
 
 def test_cached_dependency_pinned_unpinned_resolution_multi_run():
     # type: () -> None
-    # This exercises the issue described here: https://github.com/pantsbuild/pex/issues/178
+    # This exercises the issue described here: https://github.com/pex-tool/pex/issues/178
     project1_0_0 = build_wheel(name="project", version="1.0.0")
     project1_1_0 = build_wheel(name="project", version="1.1.0")
 
     with temporary_dir() as td:
         for wheel in (project1_0_0, project1_1_0):
             safe_copy(wheel, os.path.join(td, os.path.basename(wheel)))
-        with temporary_dir() as cd:
+        with temporary_dir() as cd, cache(cd):
             # First run, pinning 1.0.0 in the cache
-            installed_dists = local_resolve(
-                requirements=["project==1.0.0"], find_links=[td], cache=cd
-            )
-            assert len(installed_dists) == 1
-            assert installed_dists[0].distribution.version == "1.0.0"
+            resolved_dists = local_resolve(requirements=["project==1.0.0"], find_links=[td])
+            assert len(resolved_dists) == 1
+            assert resolved_dists[0].distribution.version == "1.0.0"
 
-            # Second, run, the unbounded 'project' req will find the 1.0.0 in the cache. But should also
-            # return SourcePackages found in td
-            installed_dists = local_resolve(requirements=["project"], find_links=[td], cache=cd)
-            assert len(installed_dists) == 1
-            assert installed_dists[0].distribution.version == "1.1.0"
+            # Second, run, the unbounded 'project' req will find the 1.0.0 in the cache. But should
+            # also return SourcePackages found in td
+            resolved_dists = local_resolve(requirements=["project"], find_links=[td])
+            assert len(resolved_dists) == 1
+            assert resolved_dists[0].distribution.version == "1.1.0"
 
 
 def test_intransitive():
@@ -163,11 +189,11 @@ def test_intransitive():
     with temporary_dir() as td:
         for wheel in (foo1_0, bar1_0):
             safe_copy(wheel, os.path.join(td, os.path.basename(wheel)))
-        with temporary_dir() as cd:
-            installed_dists = local_resolve(
-                requirements=["foo", "bar"], find_links=[td], cache=cd, transitive=False
+        with temporary_dir() as cd, cache(cd):
+            resolved_dists = local_resolve(
+                requirements=["foo", "bar"], find_links=[td], transitive=False
             )
-            assert len(installed_dists) == 2
+            assert len(resolved_dists) == 2
 
 
 def test_resolve_prereleases():
@@ -180,12 +206,12 @@ def test_resolve_prereleases():
             safe_copy(wheel, os.path.join(td, os.path.basename(wheel)))
 
         def assert_resolve(expected_version, **resolve_kwargs):
-            installed_dists = local_resolve(
+            resolved_dists = local_resolve(
                 requirements=["dep>=1,<4"], find_links=[td], **resolve_kwargs
             )
-            assert 1 == len(installed_dists)
-            installed_dist = installed_dists[0]
-            assert expected_version == installed_dist.distribution.version
+            assert 1 == len(resolved_dists)
+            resolved_dist = resolved_dists[0]
+            assert expected_version == resolved_dist.distribution.version
 
         assert_resolve("2.0.0")
         assert_resolve("2.0.0", allow_prereleases=False)
@@ -208,12 +234,12 @@ def test_resolve_extra_setup_py():
         with temporary_dir() as td:
             safe_copy(project2_wheel, os.path.join(td, os.path.basename(project2_wheel)))
 
-            installed_dists = local_resolve(
+            resolved_dists = local_resolve(
                 requirements=["{}[foo]".format(project1_dir)], find_links=[td]
             )
             assert {_parse_requirement(req) for req in ("project1==1.0.0", "project2==2.0.0")} == {
-                _parse_requirement(installed_dist.distribution.as_requirement())
-                for installed_dist in installed_dists
+                _parse_requirement(resolved_dist.distribution.as_requirement())
+                for resolved_dist in resolved_dists
             }
 
 
@@ -227,24 +253,36 @@ def test_resolve_extra_wheel():
         for wheel in (project1_wheel, project2_wheel):
             safe_copy(wheel, os.path.join(td, os.path.basename(wheel)))
 
-        installed_dists = local_resolve(requirements=["project1[foo]"], find_links=[td])
+        resolved_dists = local_resolve(requirements=["project1[foo]"], find_links=[td])
         assert {_parse_requirement(req) for req in ("project1==1.0.0", "project2==2.0.0")} == {
-            _parse_requirement(installed_dist.distribution.as_requirement())
-            for installed_dist in installed_dists
+            _parse_requirement(resolved_dist.distribution.as_requirement())
+            for resolved_dist in resolved_dists
         }
 
 
 def resolve_wheel_names(**kwargs):
     # type: (**Any) -> List[str]
     return [
-        os.path.basename(installed_distribution.distribution.location)
-        for installed_distribution in resolve(**kwargs).installed_distributions
+        os.path.basename(resolved_distribution.distribution.location)
+        for resolved_distribution in resolve(**kwargs).distributions
     ]
 
 
-def resolve_p537_wheel_names(**kwargs):
-    # type: (**Any) -> List[str]
-    return resolve_wheel_names(requirements=["p537==1.0.4"], transitive=False, **kwargs)
+def resolve_p537_wheel_names(
+    cache_dir,  # type: str
+    **kwargs  # type: Any
+):
+    # type: (...) -> List[str]
+    with cache(cache_dir):
+        return resolve_wheel_names(
+            requirements=[
+                "p537=={version}".format(
+                    version="1.0.8" if sys.version_info[:2] >= (3, 6) else "1.0.5"
+                )
+            ],
+            transitive=False,
+            **kwargs
+        )
 
 
 @pytest.fixture(scope="module")
@@ -265,11 +303,11 @@ def test_resolve_current_platform(p537_resolve_cache):
         # strings to Platform objects.
         current_platform = (None,)
         return resolve_p537_wheel_names(
-            cache=p537_resolve_cache,
+            cache_dir=p537_resolve_cache,
             targets=Targets(platforms=current_platform, interpreters=tuple(interpreters)),
         )
 
-    other_python_version = PY310 if PY_VER == (3, 7) else PY37
+    other_python_version = PY311 if PY_VER == (3, 9) else PY39
     other_python = PythonInterpreter.from_binary(ensure_python_interpreter(other_python_version))
     current_python = PythonInterpreter.get()
 
@@ -292,22 +330,22 @@ def test_resolve_current_platform(p537_resolve_cache):
 )
 def test_resolve_current_and_foreign_platforms(p537_resolve_cache):
     # type: (str) -> None
-    foreign_platform = "macosx-10.13-x86_64-cp-37-m" if IS_LINUX else "manylinux1_x86_64-cp-37-m"
+    foreign_platform = "macosx-13.0-x86_64-cp-37-m" if IS_LINUX else "manylinux1_x86_64-cp-37-m"
 
     def resolve_current_and_foreign(interpreters=()):
         # type: (Iterable[PythonInterpreter]) -> List[str]
 
         # N.B.: None stands in for the "current" platform at higher layers that parse platform
         # strings to Platform objects.
-        platforms = (None, Platform.create(foreign_platform))
+        platforms = (None, abbreviated_platforms.create(foreign_platform))
         return resolve_p537_wheel_names(
-            cache=p537_resolve_cache,
+            cache_dir=p537_resolve_cache,
             targets=Targets(platforms=platforms, interpreters=tuple(interpreters)),
         )
 
     assert 2 == len(resolve_current_and_foreign())
 
-    other_python_version = PY310 if PY_VER == (3, 7) else PY37
+    other_python_version = PY311 if PY_VER == (3, 9) else PY39
     other_python = PythonInterpreter.from_binary(ensure_python_interpreter(other_python_version))
     current_python = PythonInterpreter.get()
 
@@ -320,30 +358,33 @@ def test_resolve_current_and_foreign_platforms(p537_resolve_cache):
     assert 3 == len(resolve_current_and_foreign(interpreters=[current_python, other_python]))
 
 
-def test_resolve_foreign_abi3():
-    # type: () -> None
+def test_resolve_foreign_abi3(tmpdir):
+    # type: (Any) -> None
     # For version 2.8, cryptography publishes the following abi3 wheels for linux and macosx:
     # cryptography-2.8-cp34-abi3-macosx_10_6_intel.whl
     # cryptography-2.8-cp34-abi3-manylinux1_x86_64.whl
     # cryptography-2.8-cp34-abi3-manylinux2010_x86_64.whl
 
-    cryptogrpahy_resolve_cache = safe_mkdtemp()
+    cryptogrpahy_resolve_cache = os.path.join(str(tmpdir), "pex_root")
     foreign_ver = "37" if PY_VER == (3, 6) else "36"
 
     def resolve_cryptography_wheel_names(manylinux):
-        return resolve_wheel_names(
-            requirements=["cryptography==2.8"],
-            targets=Targets(
-                platforms=(
-                    Platform.create("linux_x86_64-cp-{}-m".format(foreign_ver)),
-                    Platform.create("macosx_10.11_x86_64-cp-{}-m".format(foreign_ver)),
+        with cache(cryptogrpahy_resolve_cache):
+            return resolve_wheel_names(
+                requirements=["cryptography==2.8"],
+                targets=Targets(
+                    platforms=(
+                        abbreviated_platforms.create(
+                            "linux_x86_64-cp-{}-m".format(foreign_ver), manylinux=manylinux
+                        ),
+                        abbreviated_platforms.create(
+                            "macosx_10.11_x86_64-cp-{}-m".format(foreign_ver), manylinux=manylinux
+                        ),
+                    ),
                 ),
-                assume_manylinux=manylinux,
-            ),
-            transitive=False,
-            build=False,
-            cache=cryptogrpahy_resolve_cache,
-        )
+                transitive=False,
+                build_configuration=BuildConfiguration.create(allow_builds=False),
+            )
 
     wheel_names = resolve_cryptography_wheel_names(manylinux="manylinux2014")
     assert {
@@ -366,29 +407,33 @@ def test_resolve_foreign_abi3():
 
 def test_issues_851():
     # type: () -> None
-    # Previously, the PY37 resolve would fail post-resolution checks for configparser, pathlib2 and
-    # contextlib2 which are only required for python_version<3.
+    # Previously, the PY38 resolve would fail post-resolution checks for importlib-metadata,
+    # configparser, pathlib2 and contextlib2 which are only required for python_version<3.
 
-    def resolve_pytest(python_version, pytest_version):
-        interpreter = PythonInterpreter.from_binary(ensure_python_interpreter(python_version))
+    def resolve_pytest(python, pytest_version):
+        interpreter = PythonInterpreter.from_binary(python)
         result = resolve(
             targets=Targets(interpreters=(interpreter,)),
             requirements=["pytest=={}".format(pytest_version)],
         )
         project_to_version = {
-            installed_dist.distribution.key: installed_dist.distribution.version
-            for installed_dist in result.installed_distributions
+            resolved_dist.distribution.project_name: resolved_dist.distribution.version
+            for resolved_dist in result.distributions
         }
         assert project_to_version["pytest"] == pytest_version
         return project_to_version
 
-    resolved_project_to_version = resolve_pytest(python_version=PY37, pytest_version="5.3.4")
-    assert "importlib-metadata" in resolved_project_to_version
+    resolved_project_to_version = resolve_pytest(
+        python=ensure_python_interpreter(PY39), pytest_version="5.3.4"
+    )
+    assert "importlib-metadata" not in resolved_project_to_version
     assert "configparser" not in resolved_project_to_version
     assert "pathlib2" not in resolved_project_to_version
     assert "contextlib2" not in resolved_project_to_version
 
-    resolved_project_to_version = resolve_pytest(python_version=PY27, pytest_version="4.6.9")
+    resolved_project_to_version = resolve_pytest(
+        python=ensure_python_interpreter(PY27), pytest_version="4.6.9"
+    )
     assert "importlib-metadata" in resolved_project_to_version
     assert "configparser" in resolved_project_to_version
     assert "pathlib2" in resolved_project_to_version
@@ -406,7 +451,7 @@ def test_issues_892():
         import sys
 
 
-        # This puts python3.8 stdlib on PYTHONPATH.
+        # This puts python3.10 stdlib on PYTHONPATH.
         os.environ['PYTHONPATH'] = os.pathsep.join(sys.path)
 
 
@@ -442,7 +487,7 @@ def test_issues_892():
     )
 
 
-def test_download():
+def test_download2():
     # type: () -> None
     project1_sdist = create_sdist(
         name="project1", version="1.0.0", extras_require={"foo": ["project2"]}
@@ -454,13 +499,14 @@ def test_download():
         install_reqs=["setuptools==44.1.0"],
     )
 
-    downloaded_by_target = defaultdict(list)
+    downloaded_by_target = defaultdict(list)  # type: DefaultDict[Target, List[Distribution]]
     result = download(
         requirements=["{}[foo]".format(project1_sdist)],
         find_links=[os.path.dirname(project2_wheel)],
+        resolver=ConfiguredResolver.default(),
     )
     for local_distribution in result.local_distributions:
-        distribution = pkginfo.get_metadata(local_distribution.path)
+        distribution = Distribution.load(local_distribution.path)
         downloaded_by_target[local_distribution.target].append(distribution)
 
     assert 1 == len(downloaded_by_target)
@@ -468,103 +514,80 @@ def test_download():
     target, distributions = downloaded_by_target.popitem()
     assert targets.current() == target
 
-    distributions_by_name = {distribution.name: distribution for distribution in distributions}
+    distributions_by_name = {
+        distribution.project_name: distribution for distribution in distributions
+    }
     assert 3 == len(distributions_by_name)
 
-    def assert_dist(project_name, dist_type, version):
-        dist = distributions_by_name[project_name]
-        assert dist_type is type(dist)
-        assert version == dist.version
-
-    assert_dist("project1", pkginfo.SDist, "1.0.0")
-    assert_dist("project2", pkginfo.Wheel, "2.0.0")
-    assert_dist("setuptools", pkginfo.Wheel, "44.1.0")
-
-
-def test_install():
-    # type: () -> None
-    project1_sdist = create_sdist(name="project1", version="1.0.0")
-    project2_wheel = build_wheel(name="project2", version="2.0.0")
-
-    installed_by_target = defaultdict(list)
-    for installed_distribution in install(
-        [LocalDistribution(path=dist) for dist in (project1_sdist, project2_wheel)]
+    def assert_dist(
+        project_name,  # type: str
+        version,  # type: str
+        is_wheel,  # type: bool
     ):
-        installed_by_target[installed_distribution.target].append(
-            installed_distribution.distribution
+        # type: (...) -> None
+
+        dist = distributions_by_name[project_name]
+        assert version == dist.version
+        assert is_wheel == (
+            dist_metadata.is_wheel(dist.location) and zipfile.is_zipfile(dist.location)
         )
 
-    assert 1 == len(installed_by_target)
-
-    target, distributions = installed_by_target.popitem()
-    assert targets.current() == target
-
-    distributions_by_name = {distribution.key: distribution for distribution in distributions}
-    assert 2 == len(distributions_by_name)
-    assert "1.0.0" == distributions_by_name["project1"].version
-    assert "2.0.0" == distributions_by_name["project2"].version
-
-    assert 2 == len(
-        {distribution.location for distribution in distributions}
-    ), "Expected installed distributions to have independent chroot paths."
-
-
-def test_install_unsatisfiable():
-    # type: () -> None
-    project1_sdist = create_sdist(name="project1", version="1.0.0")
-    project2_wheel = build_wheel(name="project2", version="2.0.0", install_reqs=["project1==1.0.1"])
-    local_distributions = [
-        LocalDistribution(path=dist) for dist in (project1_sdist, project2_wheel)
-    ]
-
-    assert 2 == len(install(local_distributions, ignore_errors=True))
-
-    with pytest.raises(Unsatisfiable):
-        install(local_distributions, ignore_errors=False)
-
-
-def test_install_invalid_local_distribution():
-    # type: () -> None
-    project1_sdist = create_sdist(name="project1", version="1.0.0")
-
-    valid_local_sdist = LocalDistribution(project1_sdist)
-    assert 1 == len(install([valid_local_sdist]))
-
-    with pytest.raises(IntegrityError):
-        install([LocalDistribution(project1_sdist, fingerprint="mismatch")])
-
-    project1_wheel = build_wheel(name="project1", version="1.0.0")
-    with pytest.raises(IntegrityError):
-        install([LocalDistribution(project1_wheel, fingerprint=valid_local_sdist.fingerprint)])
+    assert_dist("project1", "1.0.0", is_wheel=False)
+    assert_dist("project2", "2.0.0", is_wheel=True)
+    assert_dist("setuptools", "44.1.0", is_wheel=True)
 
 
+@pytest.mark.skipif(
+    sys.version_info[:2] >= (3, 12) or PipVersion.DEFAULT >= PipVersion.v24_1,
+    reason=(
+        "We need to use setuptools<66, but Python 3.12+ require greater. We also need to avoid "
+        "Pip>=24.1 which upgrades its vendored packaging to a version that rejects invalid "
+        "versions"
+    ),
+)
 def test_resolve_arbitrary_equality_issues_940():
-    # type: () -> None
+    # type: (...) -> None
+
+    def prepare_project(project_dir):
+        # type: (str) -> None
+        with open(os.path.join(project_dir, "pyproject.toml"), "w") as fp:
+            fp.write(
+                dedent(
+                    """\
+                    [build-system]
+                    # Setuptools 66 removed support for PEP-440 non-compliant versions.
+                    # See: https://setuptools.pypa.io/en/stable/history.html#v66-0-0
+                    requires = ["setuptools<66"]
+                    """
+                )
+            )
+
     dist = create_sdist(
+        prepare_project=prepare_project,
         name="foo",
         version="1.0.2-fba4511",
         python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*",
     )
-    installed_distributions = local_resolve(
+    resolved_distributions = resolve(
         requirements=[dist],
         # We need this to allow the invalid version above to sneak by pip wheel metadata
         # verification.
         verify_wheels=False,
-    )
+    ).distributions
 
-    assert len(installed_distributions) == 1
-    requirements = installed_distributions[0].direct_requirements
+    assert len(resolved_distributions) == 1
+    requirements = resolved_distributions[0].direct_requirements
     assert 1 == len(requirements), (
         "The foo requirement was direct; so the resulting resolved distribution should carry the "
         "associated requirement."
     )
-    assert [("===", "1.0.2-fba4511")] == requirements[0].specs
+    assert "===1.0.2-fba4511" == str(requirements[0].specifier)
     assert requirements[0].marker is None
 
 
 def test_resolve_overlapping_requirements_discriminated_by_markers_issues_1196(py27):
     # type: (PythonInterpreter) -> None
-    installed_distributions = resolve(
+    resolved_distributions = resolve(
         requirements=[
             "setuptools<45; python_full_version == '2.7.*'",
             "setuptools; python_version > '2.7'",
@@ -572,23 +595,23 @@ def test_resolve_overlapping_requirements_discriminated_by_markers_issues_1196(p
         targets=Targets(
             interpreters=(py27,),
         ),
-    ).installed_distributions
-    assert 1 == len(installed_distributions)
-    installed_distribution = installed_distributions[0]
-    assert 1 == len(installed_distribution.direct_requirements)
+    ).distributions
+    assert 1 == len(resolved_distributions)
+    resolved_distribution = resolved_distributions[0]
+    assert 1 == len(resolved_distribution.direct_requirements)
     assert (
         Requirement.parse("setuptools<45; python_full_version == '2.7.*'")
-        == installed_distribution.direct_requirements[0]
+        == resolved_distribution.direct_requirements[0]
     )
     assert (
         Requirement.parse("setuptools==44.1.1")
-        == installed_distribution.distribution.as_requirement()
+        == resolved_distribution.distribution.as_requirement()
     )
 
 
 def test_pip_proprietary_url_with_markers_issues_1415():
     # type: () -> None
-    installed_dists = resolve(
+    resolved_dists = resolve(
         requirements=[
             (
                 "https://files.pythonhosted.org/packages/53/18/"
@@ -597,15 +620,15 @@ def test_pip_proprietary_url_with_markers_issues_1415():
             ),
             "ansicolors==1.1.8; sys_platform == '{}'".format(sys.platform),
         ]
-    ).installed_distributions
-    assert len(installed_dists) == 1
+    ).distributions
+    assert len(resolved_dists) == 1
 
-    installed_dist = installed_dists[0]
-    assert Requirement.parse("ansicolors==1.1.8") == installed_dist.distribution.as_requirement()
-    assert 1 == len(installed_dist.direct_requirements)
+    resolved_dist = resolved_dists[0]
+    assert Requirement.parse("ansicolors==1.1.8") == resolved_dist.distribution.as_requirement()
+    assert 1 == len(resolved_dist.direct_requirements)
     assert (
         Requirement.parse("ansicolors==1.1.8; sys_platform == '{}'".format(sys.platform))
-        == installed_dist.direct_requirements[0]
+        == resolved_dist.direct_requirements[0]
     )
 
 
@@ -615,14 +638,37 @@ def test_duplicate_requirements_issues_1550():
     with pytest.raises(Unsatisfiable):
         resolve(requirements=["PyJWT", "PyJWT==1.7.1"], resolver_version=ResolverVersion.PIP_LEGACY)
 
-    installed_dists = resolve(
+    resolved_dists = resolve(
         requirements=["PyJWT", "PyJWT==1.7.1"], resolver_version=ResolverVersion.PIP_2020
     )
-    assert len(installed_dists.installed_distributions) == 1
-    installed_distribution = installed_dists.installed_distributions[0]
+    assert len(resolved_dists.distributions) == 1
+    resolved_distribution = resolved_dists.distributions[0]
     assert {Requirement.parse("PyJWT"), Requirement.parse("PyJWT==1.7.1")} == set(
-        installed_distribution.direct_requirements
+        resolved_distribution.direct_requirements
     )
-    distribution = installed_distribution.distribution
+    distribution = resolved_distribution.distribution
     assert "PyJWT" == distribution.project_name
     assert "1.7.1" == distribution.version
+
+
+def test_check_resolve_prerelease_transitive_dependencies_issue_1730(tmpdir):
+    # type: (Any) -> None
+
+    indirect_wheel = build_wheel(name="indirect", version="2.12.0.dev3")
+    direct_wheel = build_wheel(
+        name="direct", version="2.12.0.dev3", install_reqs=["indirect==2.12.0.dev3"]
+    )
+
+    find_links = os.path.join(str(tmpdir), "find-links")
+    os.mkdir(find_links)
+    for wheel in direct_wheel, indirect_wheel:
+        shutil.move(wheel, find_links)
+
+    resolved = resolve(
+        requirements=["direct==2.12.dev3"],
+        allow_prereleases=False,
+        ignore_errors=False,
+        indexes=[],
+        find_links=[find_links],
+    )
+    assert 2 == len(resolved.distributions)
